import collections
import itertools
import logging

from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions
from waldur_ansible.common.exceptions import AnsibleBackendError
from waldur_mastermind.packages import models as package_models
from waldur_openstack.openstack_tenant import serializers as openstack_serializers

from . import serializers

logger = logging.getLogger(__name__)


class InMemoryJob:
    """
    This class is used instead of database object for Ansible Job model.
    """

    def __init__(self, **kwargs):
        self.output = ''
        self.__dict__.update(kwargs)

    def get_tag(self):
        return 'VALID_JOB_TAG'

    def save(self, *args, **kwargs):
        """Skip database interaction"""


class InMemoryResource:
    """
    This class allows to treat dictionary as an object
    For example, use resource.flavor.cores instead of resource['flavor'].cores
    """

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def get_job(request):
    """
    Construct in-memory Ansible job object using default values for SSH key and name.
    """
    serializer = serializers.JobEstimateSerializer(data=request.data, context={'request': request})
    serializer.is_valid(raise_exception=True)

    return InMemoryJob(
        name='VALID_JOB_NAME',
        **serializer.validated_data
    )


def get_prices(service_settings):
    """
    Fetch prices for service settings.
    Here's diagram to illustrate relationships between involved entities:

    Job => Service project link => Service => Service settings =>
    Tenant as scope => Extra configuration dict => Package template => Package component.

    Raises a validation exception if any link is broken.
    """
    tenant = service_settings.scope
    if not tenant:
        raise exceptions.ValidationError(_('Provider is not connected to the OpenStack tenant.'))

    package_uuid = tenant.extra_configuration.get('package_uuid')
    if not package_uuid:
        raise exceptions.ValidationError(_('OpenStack tenant is not connected to the VPC package.'))

    try:
        package_template = package_models.PackageTemplate.objects.get(uuid=package_uuid)
    except package_models.PackageTemplate.DoesNotExist:
        raise exceptions.ValidationError(_('VPC package is not connected to the package template.'))

    components = package_template.components

    try:
        return {
            'cpu': components.get(type=package_models.PackageComponent.Types.CORES).price,
            'ram': components.get(type=package_models.PackageComponent.Types.RAM).price,
            'disk': components.get(type=package_models.PackageComponent.Types.STORAGE).price,
        }
    except package_models.PackageComponent.DoesNotExist:
        raise exceptions.ValidationError(_('Package template does not have required components.'))


def evaluate_job(job):
    """
    Extract JSON dictionaries for resources provisioning from Ansible playbook.
    """
    backend = job.playbook.get_backend()
    try:
        backend.run_job(job, check_mode=True)
        return backend.decode_output(job.output)
    except AnsibleBackendError as e:
        logger.debug('Unable to process Ansible playbook. '
                     'Error message is %s, job output is %s',
                     e.message, job.output)
        return []


def get_resources(request, job):
    """
    Get list of resources to be provisioned for Ansible job.
    Raises an exception if provisioning is not possible.
    """
    items = evaluate_job(job)
    errors = []
    resources = []

    settings = job.service_project_link.service.settings
    link = job.service_project_link

    for item in items:
        if 'ssh_public_key' in item:
            item['ssh_public_key'] = request.data['ssh_public_key']
        serializer = openstack_serializers.InstanceSerializer(data=item, context={'request': request})
        if not serializer.is_valid():
            errors.append(serializer.errors)
        else:
            resource = InMemoryResource(**serializer.validated_data)

            quota_errors = list(itertools.chain(
                settings.validate_quota_change({settings.Quotas.instances: 1}),
                settings.validate_quota_change({settings.Quotas.ram: resource.flavor.ram}),
                settings.validate_quota_change({settings.Quotas.vcpu: resource.flavor.cores}),
                link.validate_quota_change({link.Quotas.ram: resource.flavor.ram}),
                link.validate_quota_change({link.Quotas.vcpu: resource.flavor.cores}),
            ))
            if quota_errors:
                errors.append(_('One or more quotas were exceeded: %s') % ';'.join(quota_errors))
            else:
                resources.append(resource)

    if errors:
        raise exceptions.ValidationError(errors)
    return resources


def get_requirements(resources):
    """
    Calculate quotas requirements for CPU, RAM and disk.
    """
    requirements = collections.defaultdict(float)
    for resource in resources:
        requirements['cpu'] += resource.flavor.cores
        requirements['ram'] += resource.flavor.ram
        requirements['disk'] += resource.system_volume_size + getattr(resource, 'data_volume_size', 0)
    return requirements


def get_total_cost(requirements, prices):
    """
    Calculate total cost of provisioned resources
    with respect to requirements and package prices.
    """
    return sum(
        float(prices[component]) * float(requirements[component])
        for component in ('cpu', 'ram', 'disk')
    )


def get_report(request):
    """
    Calculate prices, requirements and total cost for resource provisioning.
    Raises an error if provisioning parameters are invalid or quota is exceeded.
    Please note that currently it is limited to OpenStack instance provisioning.
    """
    job = get_job(request)
    service_settings = job.service_project_link.service.settings
    resources = get_resources(request, job)
    requirements = get_requirements(resources)

    prices = {'cpu': 0, 'ram': 0, 'disk': 0}
    try:
        prices = get_prices(service_settings)
    except exceptions.ValidationError as e:
        logger.debug('Unable to get prices for service settings %s. Error is %s.', service_settings, e)

    cost = get_total_cost(requirements, prices)

    return {
        'prices': prices,
        'requirements': requirements,
        'cost': cost
    }



from random import randint

from django.conf import settings
from django.db import models
from django_fsm import FSMField, transition
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models


def _default_message():
    return '{0:4.0f}'.format(randint(0, 9999)).replace(' ', '0')  # nosec


class AuthResult(core_models.UuidMixin, core_models.ErrorMessageMixin, TimeStampedModel):
    class States:
        SCHEDULED = 'Scheduled'
        PROCESSING = 'Processing'
        OK = 'OK'
        CANCELED = 'Canceled'
        ERRED = 'Erred'

        CHOICES = ((SCHEDULED, SCHEDULED), (PROCESSING, PROCESSING), (OK, OK), (CANCELED, CANCELED), (ERRED, ERRED))

    user = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL, related_name='auth_valimo_results', null=True)
    phone = models.CharField(max_length=30)
    message = models.CharField(max_length=4, default=_default_message, help_text='This message will be shown to user.')
    state = FSMField(choices=States.CHOICES, default=States.SCHEDULED)
    details = models.CharField(max_length=255, blank=True, help_text='Cancellation details.')
    backend_transaction_id = models.CharField(max_length=100, blank=True)

    # for consistency with other models with state
    @property
    def human_readable_state(self):
        return self.state

    @transition(field=state, source=States.SCHEDULED, target=States.PROCESSING)
    def begin_processing(self):
        pass

    @transition(field=state, source=States.PROCESSING, target=States.OK)
    def set_ok(self):
        pass

    @transition(field=state, source=States.PROCESSING, target=States.CANCELED)
    def set_canceled(self):
        pass

    @transition(field=state, source='*', target=States.ERRED)
    def set_erred(self):
        pass



import django_filters
from django_filters.widgets import BooleanWidget

from waldur_core.core import filters as core_filters
from waldur_core.structure import filters as structure_filters
from waldur_openstack.openstack_tenant.utils import get_valid_availability_zones

from . import models


class OpenStackTenantServiceProjectLinkFilter(structure_filters.BaseServiceProjectLinkFilter):
    service = core_filters.URLFilter(view_name='openstacktenant-detail', field_name='service__uuid')

    class Meta(structure_filters.BaseServiceProjectLinkFilter.Meta):
        model = models.OpenStackTenantServiceProjectLink


class FlavorFilter(structure_filters.ServicePropertySettingsFilter):

    o = django_filters.OrderingFilter(fields=('cores', 'ram', 'disk'))

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Flavor
        fields = dict({
            'cores': ['exact', 'gte', 'lte'],
            'ram': ['exact', 'gte', 'lte'],
            'disk': ['exact', 'gte', 'lte'],
        }, **{field: ['exact'] for field in structure_filters.ServicePropertySettingsFilter.Meta.fields})


class NetworkFilter(structure_filters.ServicePropertySettingsFilter):

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Network
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('type', 'is_external')


class SubNetFilter(structure_filters.ServicePropertySettingsFilter):
    network = core_filters.URLFilter(view_name='openstacktenant-network-detail', field_name='network__uuid')
    network_uuid = django_filters.UUIDFilter(field_name='network__uuid')

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.SubNet
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('ip_version', 'enable_dhcp')


class FloatingIPFilter(structure_filters.ServicePropertySettingsFilter):
    free = django_filters.BooleanFilter(field_name='internal_ip', lookup_expr='isnull', widget=BooleanWidget)

    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.FloatingIP
        fields = structure_filters.ServicePropertySettingsFilter.Meta.fields + ('runtime_state', 'address', 'is_booked')


class VolumeFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', field_name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(field_name='instance__uuid')

    snapshot = core_filters.URLFilter(
        view_name='openstacktenant-snapshot-detail', field_name='restoration__snapshot__uuid')
    snapshot_uuid = django_filters.UUIDFilter(field_name='restoration__snapshot__uuid')

    availability_zone_name = django_filters.CharFilter(field_name='availability_zone__name')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Volume
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('instance__name', 'instance_name'),
        ('size', 'size'),
    )


class SnapshotFilter(structure_filters.BaseResourceFilter):
    source_volume_uuid = django_filters.UUIDFilter(field_name='source_volume__uuid')
    source_volume = core_filters.URLFilter(view_name='openstacktenant-volume-detail', field_name='source_volume__uuid')
    backup_uuid = django_filters.UUIDFilter(field_name='backups__uuid')
    backup = core_filters.URLFilter(view_name='openstacktenant-backup-detail', field_name='backups__uuid')

    snapshot_schedule = core_filters.URLFilter(
        view_name='openstacktenant-snapshot-schedule-detail', field_name='snapshot_schedule__uuid')
    snapshot_schedule_uuid = django_filters.UUIDFilter(field_name='snapshot_schedule__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Snapshot
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state',)

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('source_volume__name', 'source_volume_name'),
        ('size', 'size'),
    )


class InstanceAvailabilityZoneFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.InstanceAvailabilityZone


class InstanceFilter(structure_filters.BaseResourceFilter):
    external_ip = django_filters.CharFilter(field_name='internal_ips_set__floating_ips__address')
    availability_zone_name = django_filters.CharFilter(field_name='availability_zone__name')
    attach_volume_uuid = django_filters.UUIDFilter(method='filter_attach_volume')

    def filter_attach_volume(self, queryset, name, value):
        """
        This filter is used for in volume attachment dialog.
        It allows to filter out instances that could be attached to the given volume.
        """
        try:
            volume = models.Volume.objects.get(uuid=value)
        except models.Volume.DoesNotExist:
            return queryset.none()

        queryset = queryset.filter(service_project_link=volume.service_project_link)

        zones_map = get_valid_availability_zones(volume)
        if volume.availability_zone and zones_map:
            zone_names = {
                nova_zone
                for (nova_zone, cinder_zone) in zones_map.items()
                if cinder_zone == volume.availability_zone.name
            }
            nova_zones = models.InstanceAvailabilityZone.objects.filter(
                settings=volume.service_project_link.service.settings,
                name__in=zone_names,
                available=True,
            )
            queryset = queryset.filter(availability_zone__in=nova_zones)
        return queryset

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Instance
        fields = structure_filters.BaseResourceFilter.Meta.fields + ('runtime_state', 'external_ip')

    ORDERING_FIELDS = structure_filters.BaseResourceFilter.ORDERING_FIELDS + (
        ('internal_ips_set__ip4_address', 'internal_ips'),
        ('internal_ips_set__floating_ips__address', 'external_ips'),
    )


class BackupFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', field_name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(field_name='instance__uuid')
    backup_schedule = core_filters.URLFilter(
        view_name='openstacktenant-backup-schedule-detail', field_name='backup_schedule__uuid')
    backup_schedule_uuid = django_filters.UUIDFilter(field_name='backup_schedule__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.Backup


class BackupScheduleFilter(structure_filters.BaseResourceFilter):
    instance = core_filters.URLFilter(view_name='openstacktenant-instance-detail', field_name='instance__uuid')
    instance_uuid = django_filters.UUIDFilter(field_name='instance__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.BackupSchedule


class SnapshotScheduleFilter(structure_filters.BaseResourceFilter):
    source_volume = core_filters.URLFilter(view_name='openstacktenant-volume-detail', field_name='source_volume__uuid')
    source_volume_uuid = django_filters.UUIDFilter(field_name='source_volume__uuid')

    class Meta(structure_filters.BaseResourceFilter.Meta):
        model = models.SnapshotSchedule


class SecurityGroupFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.SecurityGroup


class ImageFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.Image


class VolumeTypeFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.VolumeType


class VolumeAvailabilityZoneFilter(structure_filters.ServicePropertySettingsFilter):
    class Meta(structure_filters.ServicePropertySettingsFilter.Meta):
        model = models.VolumeAvailabilityZone



# -*- coding: utf-8 -*-
from django.db import migrations, models


def get_name(item):
    if item.details.get('name'):
        return item.details.get('name')
    if item.details:
        return ', '.join(['%s: %s' % (k, v) for k, v in item.details.items()])
    if item.content_type:
        return '%s.%s' % (item.content_type.app_label, item.content_type.model)
    return ''


def init_invoice_item_name(apps, schema_editor):
    InvoiceItem = apps.get_model('invoices', 'InvoiceItem')
    for item in InvoiceItem.objects.all():
        item.name = get_name(item)
        item.save(update_fields=['name'])


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0031_rename_invoice_item_model'),
    ]

    operations = [
        migrations.AddField(
            model_name='InvoiceItem',
            name='name',
            field=models.TextField(default=''),
        ),
        migrations.RunPython(init_invoice_item_name),
    ]



# Generated by Django 2.2.7 on 2019-12-03 11:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0001_squashed_0037'),
    ]

    operations = [
        migrations.AlterField(
            model_name='issue',
            name='caller',
            field=models.ForeignKey(blank=True, help_text='Waldur user who has reported the issue.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='created_issues', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 2.2.7 on 2019-11-21 22:16

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0004_add_scope_for_offering_component'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='plugin_options',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, help_text='Data used by specific plugin, such as credentials and hooks.'),
        ),
    ]



import logging
from smtplib import SMTPException

from celery import shared_task
from celery.task import Task as CeleryTask
from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import get_template
from django.template import Template, Context

from waldur_core.core import utils as core_utils

from . import backend, models

logger = logging.getLogger(__name__)


class SupportUserPullTask(CeleryTask):
    """
    Pull support users from backend.
    Note that support users are not deleted in JIRA.
    Instead, they are marked as disabled.
    Therefore, Waldur replicates the same behaviour.
    """
    name = 'support.SupportUserPullTask'

    def run(self):
        if not settings.WALDUR_SUPPORT['ENABLED']:
            return

        backend_users = backend.get_active_backend().get_users()
        for backend_user in backend_users:
            user, created = models.SupportUser.objects.get_or_create(
                backend_id=backend_user.backend_id, defaults={'name': backend_user.name})
            if not created and user.name != backend_user.name:
                user.name = backend_user.name
                user.save()
            if not user.is_active:
                user.is_active = True
                user.save()
        models.SupportUser.objects.exclude(backend_id__in=[u.backend_id for u in backend_users])\
            .update(is_active=False)


@shared_task(name='waldur_mastermind.support.pull_priorities')
def pull_priorities():
    if not settings.WALDUR_SUPPORT['ENABLED']:
        return

    backend.get_active_backend().pull_priorities()


@shared_task(name='waldur_mastermind.support.create_issue')
def create_issue(serialized_issue):
    issue = core_utils.deserialize_instance(serialized_issue)
    try:
        backend.get_active_backend().create_issue(issue)
    except Exception as e:
        issue.error_message = str(e)
        issue.save(update_fields=['error_message'])
    else:
        issue.error_message = ''
        issue.save(update_fields=['error_message'])


@shared_task(name='waldur_mastermind.support.send_issue_updated_notification')
def send_issue_updated_notification(serialized_issue, changed):
    issue = core_utils.deserialize_instance(serialized_issue)

    _send_issue_notification(
        issue=issue,
        template='issue_updated',
        extra_context={'changed': changed},
    )


@shared_task(name='waldur_mastermind.support.send_comment_added_notification')
def send_comment_added_notification(serialized_comment):
    comment = core_utils.deserialize_instance(serialized_comment)

    _send_issue_notification(
        issue=comment.issue,
        template='comment_added',
        extra_context={'comment': comment},
    )


@shared_task(name='waldur_mastermind.support.send_comment_updated_notification')
def send_comment_updated_notification(serialized_comment, old_description):
    comment = core_utils.deserialize_instance(serialized_comment)

    _send_issue_notification(
        issue=comment.issue,
        template='comment_updated',
        extra_context={
            'comment': comment,
            'old_description': old_description,
        },
    )


def _send_issue_notification(issue, template, receiver=None, extra_context=None):
    if not settings.WALDUR_SUPPORT['ENABLED']:
        return

    if settings.SUPPRESS_NOTIFICATION_EMAILS:
        message = ('Issue notifications are suppressed. '
                   'Please set SUPPRESS_NOTIFICATION_EMAILS to False to send notifications.')
        logger.info(message)
        return

    if not receiver:
        receiver = issue.caller

    context = {
        'issue_url': settings.ISSUE_LINK_TEMPLATE.format(uuid=issue.uuid),
        'site_name': settings.WALDUR_CORE['SITE_NAME'],
        'issue': issue,
    }

    if extra_context:
        context.update(extra_context)

    try:
        notification_template = models.TemplateStatusNotification.objects.get(status=issue.status)
        html_template = Template(notification_template.html)
        text_template = Template(notification_template.text)
        subject_template = Template(notification_template.subject)
    except models.TemplateStatusNotification.DoesNotExist:
        html_template = get_template('support/notification_%s.html' % template).template
        text_template = get_template('support/notification_%s.txt' % template).template
        subject_template = get_template('support/notification_%s_subject.txt' % template).template

    html_message = html_template.render(Context(context))
    text_message = text_template.render(Context(context, autoescape=False))
    subject = subject_template.render(Context(context, autoescape=False)).strip()

    logger.debug('About to send an issue update notification to %s' % receiver.email)

    try:
        send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [receiver.email], html_message=html_message)
    except SMTPException as e:
        message = 'Failed to notify a user about an issue update. Issue uuid: %s. Error: %s' % (issue.uuid.hex, e.message)
        logger.warning(message)



from . import views


def register_in(router):
    router.register(r'openstack', views.OpenStackServiceViewSet, basename='openstack')
    router.register(r'openstack-images', views.ImageViewSet, basename='openstack-image')
    router.register(r'openstack-flavors', views.FlavorViewSet, basename='openstack-flavor')
    router.register(r'openstack-volume-types', views.VolumeTypeViewSet, basename='openstack-volume-type')
    router.register(r'openstack-tenants', views.TenantViewSet, basename='openstack-tenant')
    router.register(r'openstack-service-project-link', views.OpenStackServiceProjectLinkViewSet, basename='openstack-spl')
    router.register(r'openstack-security-groups', views.SecurityGroupViewSet, basename='openstack-sgp')
    router.register(r'openstack-floating-ips', views.FloatingIPViewSet, basename='openstack-fip')
    router.register(r'openstack-networks', views.NetworkViewSet, basename='openstack-network')
    router.register(r'openstack-subnets', views.SubNetViewSet, basename='openstack-subnet')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:02
from django.db import migrations
import waldur_core.core.fields
import waldur_core.cost_tracking.models


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0001_squashed_0026'),
    ]

    operations = [
        migrations.AlterField(
            model_name='consumptiondetails',
            name='configuration',
            field=waldur_core.cost_tracking.models.ConsumableItemsField(default=dict, help_text='Current resource configuration.'),
        ),
        migrations.AlterField(
            model_name='consumptiondetails',
            name='consumed_before_update',
            field=waldur_core.cost_tracking.models.ConsumableItemsField(default=dict, help_text='How many consumables were used by resource before last update.'),
        ),
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='metadata',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, help_text='Details of the item, that corresponds price list item. Example: details of flavor.'),
        ),
        migrations.AlterField(
            model_name='priceestimate',
            name='details',
            field=waldur_core.core.fields.JSONField(default=dict, help_text='Saved scope details. Field is populated on scope deletion.'),
        ),
    ]



from django.conf import settings
from django.forms import model_to_dict
from rest_framework.authtoken.models import Token

from waldur_core.core.log import event_logger
from waldur_core.core.models import StateMixin


def create_auth_token(sender, instance, created=False, **kwargs):
    if created:
        Token.objects.create(user=instance)


def preserve_fields_before_update(sender, instance, **kwargs):
    if instance.pk is None:
        return

    meta = instance._meta
    old_instance = meta.model._default_manager.get(pk=instance.pk)

    excluded_fields = [field.name for field in meta.many_to_many]
    excluded_fields.append(meta.pk.name)
    old_values = model_to_dict(old_instance, exclude=excluded_fields)

    setattr(instance, '_old_values', old_values)


def delete_error_message(sender, instance, name, source, target, **kwargs):
    """ Delete error message if instance state changed from erred """
    if source != StateMixin.States.ERRED:
        return
    instance.error_message = ''
    instance.save(update_fields=['error_message'])


def set_default_token_lifetime(sender, instance, created=False, **kwargs):
    if created:
        # if settings used directly in model - django creates new migration every time settings change
        # Therefore - set default token_lifetime value in handler.
        if settings.WALDUR_CORE['TOKEN_LIFETIME']:
            seconds = settings.WALDUR_CORE['TOKEN_LIFETIME'].total_seconds()
            instance.token_lifetime = int(seconds)
            instance.save(update_fields=['token_lifetime'])


def log_user_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.user.info(
            'User {affected_user_username} has been created.',
            event_type='user_creation_succeeded',
            event_context={'affected_user': instance})
    else:
        old_values = instance._old_values

        password_changed = instance.password != old_values['password']
        activation_changed = instance.is_active != old_values['is_active']
        token_lifetime_changed = instance.token_lifetime != old_values['token_lifetime']
        user_updated = any(
            old_value != getattr(instance, field_name)
            for field_name, old_value in old_values.items()
            if field_name not in ('password', 'is_active', 'last_login')
        )

        if password_changed:
            event_logger.user.info(
                'Password has been changed for user {affected_user_username}.',
                event_type='user_password_updated',
                event_context={'affected_user': instance})

        if activation_changed:
            if instance.is_active:
                event_logger.user.info(
                    'User {affected_user_username} has been activated.',
                    event_type='user_activated',
                    event_context={'affected_user': instance})
            else:
                event_logger.user.info(
                    'User {affected_user_username} has been deactivated.',
                    event_type='user_deactivated',
                    event_context={'affected_user': instance})

        if token_lifetime_changed:
            event_logger.user.info(
                'Token lifetime has been changed for {affected_user_username} to {affected_user_token_lifetime}',
                event_type='user_token_lifetime_updated',
                event_context={'affected_user': instance})

        if user_updated:
            event_logger.user.info(
                'User {affected_user_username} has been updated.',
                event_type='user_update_succeeded',
                event_context={'affected_user': instance})


def log_user_delete(sender, instance, **kwargs):
    event_logger.user.info(
        'User {affected_user_username} has been deleted.',
        event_type='user_deletion_succeeded',
        event_context={'affected_user': instance})


def log_ssh_key_save(sender, instance, created=False, **kwargs):
    if created:
        event_logger.sshkey.info(
            'SSH key {ssh_key_name} has been created for user%s with username {user_username}.' % (
                ' {user_full_name}' if instance.user.full_name else ''
            ),
            event_type='ssh_key_creation_succeeded',
            event_context={'ssh_key': instance, 'user': instance.user})


def log_ssh_key_delete(sender, instance, **kwargs):
    event_logger.sshkey.info(
        'SSH key {ssh_key_name} has been deleted for user%s with username {user_username}.' % (
            ' {user_full_name}' if instance.user.full_name else ''
        ),
        event_type='ssh_key_deletion_succeeded',
        event_context={'ssh_key': instance, 'user': instance.user})


def log_token_create(sender, instance, created=False, **kwargs):
    if created:
        event_logger.token.info(
            'Token has been updated for {affected_user_username}',
            event_type='token_created',
            event_context={'affected_user': instance.user})



# Generated by Django 2.2.7 on 2020-01-21 10:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0009_offering_parent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='categorycolumn',
            name='widget',
            field=models.CharField(blank=True, help_text='Widget field allows to customise table cell rendering.', max_length=255, null=True),
        ),
    ]



from waldur_core.core import WaldurExtension


class BillingExtension(WaldurExtension):
    @staticmethod
    def django_app():
        return 'waldur_mastermind.billing'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def is_assembly():
        return True



# Generated by Django 2.2.7 on 2019-12-09 07:26

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0009_project_is_removed'),
        ('waldur_rancher', '0006_node_initial_data'),
    ]

    operations = [
        migrations.AddField(
            model_name='cluster',
            name='tenant_settings',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='structure.ServiceSettings'),
        ),
    ]



from waldur_core.core import WaldurExtension


class ZabbixExtension(WaldurExtension):

    class Settings:
        WALDUR_ZABBIX = {
            'SMS_SETTINGS': {
                # configurations for default SMS notifications.
                'SMS_EMAIL_FROM': None,
                'SMS_EMAIL_RCPT': None,
            },
            'TRIGGER_FIELDS': (
                # matching trigger object fields and TriggerResponseSerializer fields
                # https://www.zabbix.com/documentation/3.4/manual/api/reference/trigger/object
                # (serializer field name, trigger object field, serializer field type)
                ('backend_id', 'triggerid', 'ReadOnlyField'),
                ('last_change', 'lastchange', 'IntegerField'),
                ('priority', 'priority', 'IntegerField'),
                ('description', 'description', 'ReadOnlyField'),
                ('expression', 'expression', 'ReadOnlyField'),
                ('comments', 'comments', 'ReadOnlyField'),
                ('error', 'error', 'ReadOnlyField'),
                ('value', 'value', 'IntegerField'),
            )
        }

    @staticmethod
    def django_app():
        return 'waldur_zabbix'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from datetime import timedelta
        return {
            'update-monthly-slas': {
                'task': 'waldur_core.zabbix.update_sla',
                'schedule': timedelta(minutes=5),
                'args': ('monthly',),
            },
            'update-yearly-slas': {
                'task': 'waldur_core.zabbix.update_sla',
                'schedule': timedelta(minutes=10),
                'args': ('yearly',),
            },
            'update-monitoring-items': {
                'task': 'waldur_core.zabbix.update_monitoring_items',
                'schedule': timedelta(minutes=10),
            },
            'pull-zabbix-hosts': {
                'task': 'waldur_core.zabbix.pull_hosts',
                'schedule': timedelta(minutes=30),
            },
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-21 13:24
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0010_issue_resolution_sla'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='attachment',
            unique_together=set([('issue', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='comment',
            unique_together=set([('issue', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='issue',
            unique_together=set([('project', 'backend_id')]),
        ),
    ]



from waldur_core.core import WaldurExtension


class VMwareExtension(WaldurExtension):

    class Settings:
        WALDUR_VMWARE = {
            'BASIC_MODE': False,
        }

    @staticmethod
    def get_public_settings():
        return ['BASIC_MODE']

    @staticmethod
    def django_app():
        return 'waldur_vmware'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-10-15 13:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_rancher', '0002_cluster_backend_id_not_null'),
    ]

    operations = [
        migrations.AddField(
            model_name='node',
            name='controlplane_role',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='node',
            name='etcd_role',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='node',
            name='worker_role',
            field=models.BooleanField(default=False),
        ),
    ]



import functools
import logging
import operator

from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext_lazy as _
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import decorators, response, status
from rest_framework.exceptions import ValidationError

from waldur_core.core import validators as core_validators
from waldur_core.core import views as core_views
from waldur_core.structure import filters as structure_filters
from waldur_core.structure import views as structure_views
from waldur_core.structure import permissions as structure_permissions
from waldur_core.structure.managers import filter_queryset_for_user
from waldur_core.structure.models import ServiceSettings
from waldur_core.structure.permissions import is_administrator
from waldur_rancher.apps import RancherConfig

from . import models, serializers, filters, executors, exceptions, validators

logger = logging.getLogger(__name__)


class ServiceViewSet(structure_views.BaseServiceViewSet):
    queryset = models.RancherService.objects.all()
    serializer_class = serializers.ServiceSerializer


class ServiceProjectLinkViewSet(structure_views.BaseServiceProjectLinkViewSet):
    queryset = models.RancherServiceProjectLink.objects.all()
    serializer_class = serializers.ServiceProjectLinkSerializer
    filterset_class = filters.ServiceProjectLinkFilter


class ClusterViewSet(structure_views.ImportableResourceViewSet):
    queryset = models.Cluster.objects.all()
    serializer_class = serializers.ClusterSerializer
    filterset_class = filters.ClusterFilter
    update_executor = executors.ClusterUpdateExecutor

    def perform_create(self, serializer):
        cluster = serializer.save()
        user = self.request.user
        nodes = serializer.validated_data.get('node_set')

        for node_data in nodes:
            node_data['cluster'] = cluster
            models.Node.objects.create(**node_data)

        transaction.on_commit(lambda: executors.ClusterCreateExecutor.execute(
            cluster,
            user=user,
            is_heavy_task=True,
        ))

    def destroy(self, request, *args, **kwargs):
        user = self.request.user
        instance = self.get_object()
        executors.ClusterDeleteExecutor.execute(
            instance,
            user=user,
            is_heavy_task=True,
        )
        return response.Response(
            {'detail': _('Deletion was scheduled.')}, status=status.HTTP_202_ACCEPTED)

    update_validators = partial_update_validators = [
        core_validators.StateValidator(models.Cluster.States.OK),
    ]
    destroy_validators = structure_views.ImportableResourceViewSet.destroy_validators + [
        validators.all_cluster_related_vms_can_be_deleted,
    ]
    importable_resources_backend_method = 'get_clusters_for_import'
    importable_resources_serializer_class = serializers.ClusterImportableSerializer
    import_resource_serializer_class = serializers.ClusterImportSerializer
    pull_executor = executors.ClusterPullExecutor

    @decorators.action(detail=True, methods=['get'])
    def kubeconfig_file(self, request, uuid=None):
        cluster = self.get_object()
        backend = cluster.get_backend()
        try:
            config = backend.get_kubeconfig_file(cluster)
        except exceptions.RancherException:
            raise ValidationError('Unable to get kubeconfig file.')

        return response.Response({'config': config}, status=status.HTTP_200_OK)

    kubeconfig_file_validators = [core_validators.StateValidator(models.Cluster.States.OK)]


class NodeViewSet(core_views.ActionsViewSet):
    queryset = models.Node.objects.all()
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    serializer_class = serializers.NodeSerializer
    create_serializer_class = serializers.CreateNodeSerializer
    filterset_class = filters.NodeFilter
    lookup_field = 'uuid'
    disabled_actions = ['update', 'partial_update']
    create_permissions = [structure_permissions.is_staff]
    destroy_validators = [
        validators.related_vm_can_be_deleted,
    ]

    def perform_create(self, serializer):
        node = serializer.save()
        user = self.request.user
        transaction.on_commit(lambda: executors.NodeCreateExecutor.execute(
            node,
            user=user,
            is_heavy_task=True,
        ))

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        user = self.request.user
        executors.NodeDeleteExecutor.execute(
            instance,
            user=user,
            is_heavy_task=True,
        )
        return response.Response(status=status.HTTP_202_ACCEPTED)

    @decorators.action(detail=True, methods=['post'])
    def link_openstack(self, request, uuid=None):
        node = self.get_object()
        if node.content_type and node.object_id:
            raise ValidationError('Node is already linked.')
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        instance = serializer.validated_data['instance']
        node.content_type = ContentType.objects.get_for_model(instance)
        node.object_id = instance.id
        node.name = instance.name
        node.save()
        return response.Response(status=status.HTTP_200_OK)

    link_openstack_permissions = [structure_permissions.is_staff]
    link_openstack_serializer_class = serializers.LinkOpenstackSerializer

    @decorators.action(detail=True, methods=['post'])
    def unlink_openstack(self, request, uuid=None):
        node = self.get_object()
        if not node.content_type or not node.object_id:
            raise ValidationError('Node is not linked to any OpenStack instance yet.')
        node.content_type = None
        node.object_id = None
        node.save()
        return response.Response(status=status.HTTP_200_OK)

    unlink_openstack_permissions = [structure_permissions.is_staff]


class CatalogViewSet(core_views.ActionsViewSet):
    queryset = models.Catalog.objects.all()
    serializer_class = serializers.CatalogSerializer
    lookup_field = 'uuid'

    def get_queryset(self):
        settings_uuid = self.request.query_params.get('settings_uuid')
        cluster_uuid = self.request.query_params.get('cluster_uuid')
        if settings_uuid:
            return self.filter_catalogs_for_settings(settings_uuid)
        elif cluster_uuid:
            return self.filter_catalogs_for_cluster(cluster_uuid)
        else:
            return self.filter_visible_catalogs()

    def filter_catalogs_for_settings(self, settings_uuid):
        qs = ServiceSettings.objects.filter(type=RancherConfig.service_name)
        scope = get_object_or_404(qs, uuid=settings_uuid)
        ctype = ContentType.objects.get_for_model(ServiceSettings)
        return self.queryset.filter(content_type=ctype, object_id=scope.id)

    def filter_catalogs_for_cluster(self, cluster_uuid):
        qs = filter_queryset_for_user(
            queryset=models.Cluster.objects.all(),
            user=self.request.user,
        )
        cluster = get_object_or_404(qs, uuid=cluster_uuid)
        return self.queryset.filter(
            Q(content_type=ContentType.objects.get_for_model(models.Cluster),
              object_id=cluster.id) |
            Q(content_type=ContentType.objects.get_for_model(ServiceSettings),
              object_id=cluster.service_project_link.service.settings.id)
        )

    def filter_visible_catalogs(self):
        settings_subquery = self.get_filtered_subquery(
            models.ServiceSettings.objects.filter(type=RancherConfig.service_name))
        clusters_subquery = self.get_filtered_subquery(models.Cluster.objects.all())
        # TODO: Implement project-level catalogs
        visible_scopes = settings_subquery | clusters_subquery
        return self.queryset.filter(visible_scopes)

    def get_filtered_subquery(self, queryset):
        ids = filter_queryset_for_user(
            queryset=queryset,
            user=self.request.user,
        ).values_list('id', flat=True)
        content_type = ContentType.objects.get_for_model(queryset.model)
        return functools.reduce(operator.or_, [
            Q(content_type=content_type, object_id=object_id)
            for object_id in ids
        ])

    @decorators.action(detail=True, methods=['post'])
    def refresh(self, request, uuid=None):
        catalog = self.get_object()
        backend = catalog.get_backend()
        backend.refresh_catalog(catalog)
        return response.Response(status=status.HTTP_200_OK)

    refresh_permissions = [structure_permissions.is_staff]

    def perform_create(self, serializer):
        scope = serializer.validated_data['scope']
        self.check_catalog_permissions(scope)

        if isinstance(scope, ServiceSettings):
            if scope.type != RancherConfig.service_name:
                raise ValidationError('Invalid provider detected.')
        elif not isinstance(scope, models.Cluster):
            # TODO: Implement project-level catalog
            raise ValidationError('Invalid scope provided.')

        catalog = serializer.save()
        backend = catalog.get_backend()
        backend.create_catalog(catalog)

    create_serializer_class = serializers.CatalogCreateSerializer

    def perform_update(self, serializer):
        scope = serializer.instance.scope
        self.check_catalog_permissions(scope)
        catalog = serializer.save()
        backend = catalog.get_backend()
        backend.update_catalog(catalog)

    update_serializer_class = serializers.CatalogUpdateSerializer

    def perform_destroy(self, catalog):
        self.check_catalog_permissions(catalog.scope)
        backend = catalog.get_backend()
        backend.delete_catalog(catalog)
        catalog.delete()

    def check_catalog_permissions(self, scope):
        if isinstance(scope, ServiceSettings) and not self.request.user.is_staff:
            raise ValidationError(_('Only staff is allowed to manage global catalogs.'))
        if isinstance(scope, models.Cluster):
            is_administrator(self.request.user, scope.service_project_link.project)



from django.db import models


class DataciteMixin(models.Model):
    class Meta:
        abstract = True

    datacite_doi = models.CharField(max_length=255, blank=True)

    def get_datacite_title(self):
        raise NotImplementedError

    def get_datacite_creators_name(self):
        raise NotImplementedError

    def get_datacite_description(self):
        raise NotImplementedError

    def get_datacite_publication_year(self):
        raise NotImplementedError

    def get_datacite_url(self):
        raise NotImplementedError



import logging

from django.core.management.base import BaseCommand

from waldur_rancher.utils import SyncUser

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = """Sync users from Waldur to Rancher."""

    def handle(self, *args, **options):
        def print_message(count, action):
            if count == 1:
                self.stdout.write(self.style.SUCCESS('%s user has been %s.' % (count, action)))
            else:
                self.stdout.write(self.style.SUCCESS('%s users have been %s.' % (count, action)))

        result = SyncUser.run()

        for action in ['blocked', 'created', 'activated', 'updated']:
            print_message(result[action], action)



from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_ansible.playbook_jobs import models
from waldur_core.core import models as core_models
from waldur_core.core import utils as core_utils
from waldur_core.structure.serializers import PermissionFieldFilteringMixin
from waldur_openstack.openstack_tenant import models as openstack_models


class JobEstimateSerializer(PermissionFieldFilteringMixin,
                            serializers.HyperlinkedModelSerializer):
    ssh_public_key = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name='sshpublickey-detail',
        queryset=core_models.SshPublicKey.objects.all(),
        required=True,
    )
    service_project_link = serializers.HyperlinkedRelatedField(
        lookup_field='pk',
        view_name='openstacktenant-spl-detail',
        queryset=openstack_models.OpenStackTenantServiceProjectLink.objects.all(),
    )
    playbook = serializers.HyperlinkedRelatedField(
        lookup_field='uuid',
        view_name=core_utils.get_detail_view_name(models.Playbook),
        queryset=models.Playbook.objects.all(),
    )
    arguments = serializers.JSONField(default=dict)

    class Meta:
        model = models.Job
        fields = ('ssh_public_key', 'service_project_link', 'playbook', 'arguments')

    def get_filtered_field_names(self):
        return 'service_project_link',

    def check_subnet(self, attrs):
        if not self.instance:
            settings = attrs['service_project_link'].service.settings
            if not openstack_models.SubNet.objects.filter(settings=settings).exists():
                raise serializers.ValidationError(_('Selected OpenStack provider does not have any subnet yet.'))
            else:
                attrs['subnet'] = openstack_models.SubNet.objects.filter(settings=settings).first()

    def validate(self, attrs):
        if not self.instance:
            attrs['user'] = self.context['request'].user

        self.check_subnet(attrs)
        return attrs



# Generated by Django 2.2.7 on 2019-11-22 17:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0005_offering_plugin_options'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderitem',
            name='output',
            field=models.TextField(blank=True),
        ),
    ]



from rest_framework import serializers


class OrderItemSerializer(serializers.Serializer):
    attributes = serializers.ReadOnlyField(source='attributes')
    limits = serializers.ReadOnlyField(source='limits')
    project_uuid = serializers.ReadOnlyField(source='order.project.uuid')
    project_name = serializers.ReadOnlyField(source='order.project.name')
    customer_uuid = serializers.ReadOnlyField(source='order.project.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='order.project.customer.name')
    offering_uuid = serializers.ReadOnlyField(source='offering.uuid')
    offering_name = serializers.ReadOnlyField(source='offering.name')
    plan_uuid = serializers.ReadOnlyField(source='plan.uuid')
    plan_name = serializers.ReadOnlyField(source='plan.name')
    resource_uuid = serializers.ReadOnlyField(source='resource.uuid')
    resource_name = serializers.ReadOnlyField(source='resource.name')


class ResourceSerializer(serializers.Serializer):
    attributes = serializers.ReadOnlyField(source='attributes')
    limits = serializers.ReadOnlyField(source='limits')
    project_uuid = serializers.ReadOnlyField(source='project.uuid')
    project_name = serializers.ReadOnlyField(source='project.name')
    customer_uuid = serializers.ReadOnlyField(source='project.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='project.customer.name')
    offering_uuid = serializers.ReadOnlyField(source='offering.uuid')
    offering_name = serializers.ReadOnlyField(source='offering.name')
    plan_uuid = serializers.ReadOnlyField(source='plan.uuid')
    plan_name = serializers.ReadOnlyField(source='plan.name')
    resource_uuid = serializers.ReadOnlyField(source='uuid')
    resource_name = serializers.ReadOnlyField(source='name')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-09-11 14:40
import django.core.validators
from django.db import migrations, models
import re


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_remove_organization'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='username',
            field=models.CharField(help_text='Required. 128 characters or fewer. Letters, numbers and @/./+/-/_ characters', max_length=128, unique=True, validators=[django.core.validators.RegexValidator(re.compile('^[\\w.@+-]+$'), 'Enter a valid username.', 'invalid')], verbose_name='username'),
        ),
    ]



from waldur_core.logging.loggers import EventLogger, event_logger


class OpenStackPackageLogger(EventLogger):
    tenant = 'openstack.Tenant'
    service_settings = 'structure.ServiceSettings'
    package_template_name = str

    class Meta:
        event_types = (
            'openstack_package_created',
            'openstack_package_change_scheduled',
            'openstack_package_change_succeeded',
            'openstack_package_change_failed',
            'openstack_package_deleted')
        event_groups = {
            'customers': event_types,
            'packages': event_types,
            'debug_only': event_types
        }
        nullable_fields = ('service_settings',)

    @staticmethod
    def get_scopes(event_context):
        tenant = event_context['tenant']
        project = tenant.service_project_link.project
        return {tenant, project, project.customer}


event_logger.register('openstack_package', OpenStackPackageLogger)



import os

from django.db import models


def get_upload_path(instance, filename):
    path = '%s/%s' % (instance._meta.model_name, instance.uuid.hex)
    _, ext = os.path.splitext(filename)
    return '%s%s' % (path, ext)


class ImageModelMixin(models.Model):
    class Meta:
        abstract = True

    image = models.ImageField(upload_to=get_upload_path, null=True, blank=True)



import logging

import requests

from .exceptions import RancherException

logger = logging.getLogger(__name__)


class GlobalRoleId:
    admin = 'admin'
    authn_manage = 'authn-manage'
    catalogs_manage = 'catalogs-manage'
    catalogs_use = 'catalogs-use'
    clusters_create = 'clusters-create'
    kontainerdrivers_manage = 'kontainerdrivers-manage'
    nodedrivers_manage = 'nodedrivers-manage'
    podsecuritypolicytemplates_manage = 'podsecuritypolicytemplates-manage'
    roles_manage = 'roles-manage'
    settings_manage = 'settings-manage'
    user = 'user'
    user_base = 'user-base'
    users_manage = 'users-manage'


class ClusterRoleId:
    admin = 'admin'
    backups_manage = 'backups-manage'
    cluster_admin = 'cluster-admin'
    cluster_member = 'cluster-member'
    cluster_owner = 'cluster-owner'
    clustercatalogs_manage = 'clustercatalogs-manage'
    clustercatalogs_view = 'clustercatalogs-view'
    clusterroletemplatebindings_manage = 'clusterroletemplatebindings-manage'
    clusterroletemplatebindings_view = 'clusterroletemplatebindings-view'
    configmaps_manage = 'configmaps-manage'
    configmaps_view = 'configmaps-view'
    create_ns = 'create-ns'
    edit = 'edit'
    ingress_manage = 'ingress-manage'
    ingress_view = 'ingress-view'
    nodes_manage = 'nodes-manage'
    nodes_view = 'nodes-view'
    persistentvolumeclaims_manage = 'persistentvolumeclaims-manage'
    persistentvolumeclaims_view = 'persistentvolumeclaims-view'
    project_member = 'project-member'
    project_monitoring_readonly = 'project-monitoring-readonly'
    project_owner = 'project-owner'
    projectcatalogs_manage = 'projectcatalogs-manage'
    projectcatalogs_view = 'projectcatalogs-view'
    projectroletemplatebindings_manage = 'projectroletemplatebindings-manage'
    projectroletemplatebindings_view = 'projectroletemplatebindings-view'
    projects_create = 'projects-create'
    projects_view = 'projects-view'
    read_only = 'read-only'
    secrets_manage = 'secrets-manage'
    secrets_view = 'secrets-view'
    serviceaccounts_manage = 'serviceaccounts-manage'
    serviceaccounts_view = 'serviceaccounts-view'
    services_manage = 'services-manage'
    services_view = 'services-view'
    storage_manage = 'storage-manage'
    view = 'view'
    workloads_manage = 'workloads-manage'
    workloads_view = 'workloads-view'


class ProjectRoleId:
    admin = 'admin'
    backups_manage = 'backups-manage'
    cluster_admin = 'cluster-admin'
    cluster_member = 'cluster-member'
    cluster_owner = 'cluster-owner'
    clustercatalogs_manage = 'clustercatalogs-manage'
    clustercatalogs_view = 'clustercatalogs-view'
    clusterroletemplatebindings_manage = 'clusterroletemplatebindings-manage'
    clusterroletemplatebindings_view = 'clusterroletemplatebindings-view'
    configmaps_manage = 'configmaps-manage'
    configmaps_view = 'configmaps-view'
    create_ns = 'create-ns'
    edit = 'edit'
    ingress_manage = 'ingress-manage'
    ingress_view = 'ingress-view'
    nodes_manage = 'nodes-manage'
    nodes_view = 'nodes-view'
    persistentvolumeclaims_manage = 'persistentvolumeclaims-manage'
    persistentvolumeclaims_view = 'persistentvolumeclaims-view'
    project_member = 'project-member'
    project_monitoring_readonly = 'project-monitoring-readonly'
    project_owner = 'project-owner'
    projectcatalogs_manage = 'projectcatalogs-manage'
    projectcatalogs_view = 'projectcatalogs-view'
    projectroletemplatebindings_manage = 'projectroletemplatebindings-manage'
    projectroletemplatebindings_view = 'projectroletemplatebindings-view'
    projects_create = 'projects-create'
    projects_view = 'projects-view'
    read_only = 'read-only'
    secrets_manage = 'secrets-manage'
    secrets_view = 'secrets-view'
    serviceaccounts_manage = 'serviceaccounts-manage'
    serviceaccounts_view = 'serviceaccounts-view'
    services_manage = 'services-manage'
    services_view = 'services-view'
    storage_manage = 'storage-manage'
    view = 'view'
    workloads_manage = 'workloads-manage'
    workloads_view = 'workloads-view'


class RancherClient:
    """
    Rancher API client.
    See also: https://rancher.com/docs/rancher/v2.x/en/api/
    """
    def __init__(self, host, verify_ssl=True):
        """
        Initialize client with connection options.

        :param host: Rancher server IP address
        :type host: string
        :param verify_ssl: verify SSL certificates for HTTPS requests
        :type verify_ssl: bool
        """
        self._host = host
        self._base_url = '{0}/v3'.format(self._host)
        self._session = requests.Session()
        self._session.verify = verify_ssl

    def _request(self, method, endpoint, json=None, **kwargs):
        url = '%s/%s' % (self._base_url, endpoint)

        try:
            response = self._session.request(method, url, json=json, **kwargs)
        except requests.RequestException as e:
            raise RancherException(e)

        data = None
        if response.content:
            data = response.json()

        status_code = response.status_code
        if status_code in (requests.codes.ok,
                           requests.codes.created,
                           requests.codes.accepted,
                           requests.codes.no_content):
            if isinstance(data, dict) and 'value' in data:
                return data['value']
            return data
        else:
            raise RancherException(data)

    def _get(self, endpoint, **kwargs):
        return self._request('get', endpoint, **kwargs)

    def _post(self, endpoint, **kwargs):
        return self._request('post', endpoint, **kwargs)

    def _patch(self, endpoint, **kwargs):
        return self._request('patch', endpoint, **kwargs)

    def _delete(self, endpoint, **kwargs):
        return self._request('delete', endpoint, **kwargs)

    def _put(self, endpoint, **kwargs):
        return self._request('put', endpoint, **kwargs)

    def login(self, access_key, secret_key):
        """
        Login to Rancher server using access_key and secret_key.

        :param access_key: access_key to connect
        :type access_key: string
        :param secret_key: secret_key to connect
        :type secret_key: string
        :raises Unauthorized: raised if credentials are invalid.
        """
        self._post('', auth=(access_key, secret_key))
        self._session.auth = (access_key, secret_key)
        logger.debug('Successfully logged in as {0}'.format(access_key))

    def list_clusters(self):
        return self._get('clusters')['data']

    def get_cluster(self, cluster_id):
        return self._get('clusters/{0}'.format(cluster_id))

    def create_cluster(self, cluster_name):
        return self._post('clusters', json={'name': cluster_name, 'rancherKubernetesEngineConfig': {}})

    def delete_cluster(self, cluster_id):
        return self._delete('clusters/{0}'.format(cluster_id))

    def delete_node(self, node_id):
        return self._delete('nodes/{0}'.format(node_id))

    def list_cluster_registration_tokens(self):
        return self._get('clusterregistrationtokens', params={'limit': -1})['data']

    def create_cluster_registration_token(self, cluster_id):
        return self._post('clusterregistrationtoken',
                          json={'type': 'clusterRegistrationToken', 'clusterId': cluster_id})

    def get_node_command(self, cluster_id):
        cluster_list = self.list_cluster_registration_tokens()
        cluster = list(filter(lambda x: x['clusterId'] == cluster_id, cluster_list))
        if cluster:
            node_command = cluster[0]['nodeCommand']
            return node_command

    def update_cluster(self, cluster_id, new_params):
        return self._put('clusters/{0}'.format(cluster_id), json=new_params)

    def get_node(self, node_id):
        return self._get('nodes/{0}'.format(node_id))

    def get_kubeconfig_file(self, cluster_id):
        data = self._post('clusters/{0}'.format(cluster_id), params={'action': 'generateKubeconfig'})
        return data['config']

    def list_users(self):
        return self._get('users')['data']

    def create_user(self, name, username, password, mustChangePassword=True):
        return self._post('users', json={
            'name': name,
            'mustChangePassword': mustChangePassword,
            'password': password,
            'username': username,
        })

    def enable_user(self, user_id):
        return self._put('users/{0}'.format(user_id), json={
            'enabled': True,
        })

    def disable_user(self, user_id):
        return self._put('users/{0}'.format(user_id), json={
            'enabled': False,
        })

    def create_global_role(self, user_id, role):
        return self._post('globalrolebindings', json={
            'globalRoleId': role,
            'userId': user_id,
        })

    def delete_global_role(self, role_id):
        return self._delete('globalrolebindings/{0}'.format(role_id))

    def create_cluster_role(self, user_id, cluster_id, role):
        return self._post('clusterroletemplatebindings', json={
            'roleTemplateId': role,
            'clusterId': cluster_id,
            'userId': user_id,
        })

    def create_project_role(self, user_id, project_id, role):
        return self._post('projectroletemplatebindings', json={
            'roleTemplateId': role,
            'projectId': project_id,
            'userId': user_id,
        })

    def delete_user(self, user_id):
        return self._delete('users/{0}'.format(user_id))

    def delete_cluster_role(self, cluster_role_id):
        return self._delete('clusterroletemplatebindings/{0}'.format(cluster_role_id))

    def list_global_catalogs(self):
        return self._get('catalogs', params={'limit': -1})['data']

    def list_cluster_catalogs(self):
        return self._get('clustercatalogs', params={'limit': -1})['data']

    def list_project_catalogs(self):
        return self._get('projectcatalogs', params={'limit': -1})['data']

    def refresh_global_catalog(self, catalog_id):
        return self._post('catalogs/{0}'.format(catalog_id), params={'action': 'refresh'})

    def refresh_cluster_catalog(self, catalog_id):
        return self._post('clustercatalogs/{0}'.format(catalog_id), params={'action': 'refresh'})

    def refresh_project_catalog(self, catalog_id):
        return self._post('projectcatalogs/{0}'.format(catalog_id), params={'action': 'refresh'})

    def delete_global_catalog(self, catalog_id):
        return self._delete('catalogs/{0}'.format(catalog_id))

    def delete_cluster_catalog(self, catalog_id):
        return self._delete('clustercatalogs/{0}'.format(catalog_id))

    def delete_project_catalog(self, catalog_id):
        return self._delete('projectcatalogs/{0}'.format(catalog_id))

    def create_global_catalog(self, spec):
        return self._post('catalogs', json=spec)

    def create_cluster_catalog(self, spec):
        return self._post('clustercatalogs', json=spec)

    def create_project_catalog(self, spec):
        return self._post('projectcatalogs', json=spec)

    def update_global_catalog(self, catalog_id, spec):
        return self._put('catalogs/{0}'.format(catalog_id), json=spec)

    def update_cluster_catalog(self, catalog_id, spec):
        return self._put('clustercatalogs/{0}'.format(catalog_id), json=spec)

    def update_project_catalog(self, catalog_id, spec):
        return self._put('projectcatalogs/{0}'.format(catalog_id), json=spec)



from django.core.cache import cache


def is_syncing(cache_key):
    """
    This function checks if task is already running.
    """
    return cache.get(cache_key)


def renew_task_status(cache_key, timeout):
    """
    This function sets lock with timeout. Lock is valid only for 10 minutes.
    Then it should be renewed. Otherwise, lock is released.
    """
    cache.set(cache_key, True, timeout)


def release_task_status(cache_key):
    cache.set(cache_key, False)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-04-15 10:06
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_user_backend_id'),
    ]

    operations = [
        migrations.AlterField(
            model_name='user',
            name='organization',
            field=models.CharField(blank=True, max_length=255, verbose_name='organization'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-18 13:57
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0005_customer_domain'),
    ]

    operations = [
        migrations.AddField(
            model_name='customer',
            name='backend_id',
            field=models.CharField(blank=True, help_text='Organization identifier in another application.', max_length=255),
        ),
    ]



from waldur_core.core import WaldurExtension


class MarketplaceRancherExtension(WaldurExtension):
    class Settings:
        pass

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_rancher'

    @staticmethod
    def is_assembly():
        return True



from django.conf.urls import url

from waldur_core.structure import views


def register_in(router):
    router.register(r'customers', views.CustomerViewSet)
    router.register(r'project-types', views.ProjectTypeViewSet, basename='project_type')
    router.register(r'projects', views.ProjectViewSet)
    router.register(r'customer-permissions', views.CustomerPermissionViewSet, basename='customer_permission')
    router.register(r'customer-permissions-log', views.CustomerPermissionLogViewSet, basename='customer_permission_log')
    router.register(r'project-permissions', views.ProjectPermissionViewSet, basename='project_permission')
    router.register(r'project-permissions-log', views.ProjectPermissionLogViewSet, basename='project_permission_log')
    router.register(r'service-settings', views.ServiceSettingsViewSet)
    router.register(r'service-metadata', views.ServiceMetadataViewSet, basename='service_metadata')
    router.register(r'services', views.ServicesViewSet, basename='service_items')
    router.register(r'resources', views.ResourceSummaryViewSet, basename='resource')
    router.register(r'users', views.UserViewSet)
    router.register(r'keys', views.SshKeyViewSet)
    router.register(r'service-certifications', views.ServiceCertificationViewSet, basename='service-certification')
    router.register(r'divisions', views.DivisionViewSet, basename='division')


urlpatterns = [
    url(r'^stats/creation-time/$', views.CreationTimeStatsView.as_view(), name='stats_creation_time'),
    url(r'^stats/quota/$', views.AggregatedStatsView.as_view(), name='stats_quota'),
    url(r'^stats/quota/timeline/$', views.QuotaTimelineStatsView.as_view(), name='stats_quota_timeline'),
    url(r'^customers/(?P<uuid>[a-f0-9]+)/counters/$', views.CustomerCountersView.as_view({'get': 'list'}), name='customer_counters'),
    url(r'^projects/(?P<uuid>[a-f0-9]+)/counters/$', views.ProjectCountersView.as_view({'get': 'list'}), name='project_counters'),
    url(r'^user-counters/$', views.UserCountersView.as_view({'get': 'list'}), name='user_counters'),
]



from django.apps import AppConfig


class AnsibleEstimatorConfig(AppConfig):
    name = 'waldur_mastermind.ansible_estimator'
    verbose_name = 'Ansible estimator'

    def ready(self):
        pass



import logging

from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from waldur_core.core.admin import ExecutorAdminAction, UpdateOnlyModelAdmin
from waldur_core.structure import admin as structure_admin

from . import models, executors

logger = logging.getLogger(__name__)


class InvoiceItemInline(UpdateOnlyModelAdmin, admin.TabularInline):
    model = models.InvoiceItem
    readonly_fields = ('name', 'price', 'unit_price', 'unit_of_measure', 'start', 'end', 'tax', 'quantity')


class InvoiceAdmin(structure_admin.BackendModelAdmin):
    inlines = [InvoiceItemInline]
    list_display = ['customer', 'state', 'invoice_date', 'end_date', 'tax_percent', 'backend_id']
    actions = ['create_invoice', 'pull']

    class CreateInvoice(ExecutorAdminAction):
        executor = executors.InvoiceCreateExecutor
        short_description = _('Create invoice')

    create_invoice = CreateInvoice()

    class InvoicePull(ExecutorAdminAction):
        executor = executors.InvoicePullExecutor
        short_description = _('Pull invoice')

    pull = InvoicePull()


class PaymentAdmin(admin.ModelAdmin):
    list_display = ['customer', 'amount', 'state', 'backend_id']


admin.site.register(models.Invoice, InvoiceAdmin)
admin.site.register(models.Payment, PaymentAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-14 08:24
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('logging', '0005_report'),
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False)),
                ('event_type', models.CharField(db_index=True, max_length=100)),
                ('message', models.TextField()),
                ('context', django.contrib.postgres.fields.jsonb.JSONField(blank=True)),
            ],
            options={
                'abstract': False,
                'ordering': ('-created',),
            },
        ),
        migrations.CreateModel(
            name='Feed',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(db_index=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('event', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='logging.Event')),
            ],
        ),
    ]



import logging

from django.core.exceptions import ObjectDoesNotExist

from waldur_azure import models as azure_models
from waldur_core.core import utils as core_utils
from waldur_mastermind.marketplace import models as marketplace_models


logger = logging.getLogger(__name__)


def synchronize_nic(nic):
    try:
        vm = azure_models.VirtualMachine.objects.get(network_interface=nic)
        resource = marketplace_models.Resource.objects.get(scope=vm)
    except ObjectDoesNotExist:
        logger.debug('Skipping Azure virtual machine synchronization '
                     'because marketplace resource does not exist. '
                     'Resource: %s', core_utils.serialize_instance(nic))
        return
    else:
        resource.backend_metadata['internal_ips'] = vm.internal_ips
        resource.backend_metadata['external_ips'] = vm.external_ips
        resource.save(update_fields=['backend_metadata'])



# Generated by Django 2.2.7 on 2020-01-20 12:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0010_allow_null_widget'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='datacite_doi',
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



# Generated by Django 2.2.7 on 2020-01-13 16:24

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('marketplace', '0009_offering_parent'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('structure', '0009_project_is_removed'),
    ]

    operations = [
        migrations.CreateModel(
            name='Checklist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Question',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, max_length=500, verbose_name='description')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('order', models.PositiveIntegerField(default=0)),
                ('category', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='marketplace.Category')),
                ('checklist', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace_checklist.Checklist', related_name='questions',)),
            ],
            options={
                'ordering': ('checklist', 'order'),
            },
        ),
        migrations.CreateModel(
            name='Answer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('value', models.NullBooleanField()),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='structure.Project')),
                ('question', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='marketplace_checklist.Question')),
                ('user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import datetime

from django.utils.functional import cached_property

from waldur_core.core import utils as core_utils

from .base import BaseReportLine


def parse_duration(value):
    """
    Returns duration in minutes as an integer number.
    For example 00:01:00 is equal to 1
    """
    dt = datetime.datetime.strptime(value, '%H:%M:%S')
    delta = datetime.timedelta(hours=dt.hour, minutes=dt.minute, seconds=dt.second)
    return int(delta.total_seconds()) // 60


class SlurmReportLine(BaseReportLine):
    def __init__(self, line):
        self._parts = line.split('|')

    @cached_property
    def account(self):
        return self._parts[0].strip()

    @cached_property
    def user(self):
        return self._parts[3]

    @cached_property
    def cpu(self):
        return self.parse_field('cpu')

    @cached_property
    def gpu(self):
        return self.parse_field('gres/gpu')

    @cached_property
    def ram(self):
        return self.parse_field('mem')

    @cached_property
    def node(self):
        return self.parse_field('node')

    @cached_property
    def duration(self):
        return parse_duration(self._parts[2])

    @cached_property
    def _resources(self):
        pairs = self._parts[1].split(',')
        return dict(pair.split('=') for pair in pairs)

    def parse_field(self, field):
        if field not in self._resources:
            return 0
        return core_utils.parse_int(self._resources[field])



from django.db import models

from django.contrib.contenttypes import fields as ct_fields
from django.contrib.contenttypes import models as ct_models

from waldur_core.core.managers import GenericKeyMixin


class QuotaManager(GenericKeyMixin, models.Manager):
    def update_or_create_quota(self, scope, name, date, usage):
        content_type = ct_models.ContentType.objects.get_for_model(scope)
        return self.update_or_create(
            content_type=content_type,
            object_id=scope.pk,
            name=name,
            date=date,
            defaults=dict(usage=usage),
        )


class DailyQuotaHistory(models.Model):
    """
    This model stores quota usage history per day.
    It is designed to store derived data optimized for dashboard charts.
    See also related design pattern:
    https://martinfowler.com/bliki/ReportingDatabase.html
    """
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ct_models.ContentType)
    object_id = models.PositiveIntegerField()
    scope = ct_fields.GenericForeignKey('content_type', 'object_id')
    objects = QuotaManager()
    name = models.CharField(max_length=150, db_index=True)
    usage = models.BigIntegerField()
    date = models.DateField()

    class Meta:
        unique_together = ('content_type', 'object_id', 'name', 'date')



import datetime
import logging

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker
from model_utils.models import TimeStampedModel

from waldur_core.core import models as core_models, utils as core_utils
from waldur_core.core.fields import JSONField
from waldur_core.cost_tracking import managers, ConsumableItem
from waldur_core.logging.loggers import LoggableMixin
from waldur_core.structure import models as structure_models, SupportedServices

from . import CostTrackingRegister

logger = logging.getLogger(__name__)


class EstimateUpdateError(Exception):
    pass


class PriceEstimate(LoggableMixin, core_models.UuidMixin, core_models.DescendantMixin):
    """ Store prices based on both estimates and actual consumption.

        Every record holds a list of children estimates.
                   /--- Service ---\
        Customer --                 ---> SPL --> Resource
                   \--- Project ---/
        Only resource node has actual data.
        Another ones should be re-calculated on every change of resource estimate.
    """
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    scope = GenericForeignKey('content_type', 'object_id')
    details = JSONField(default=dict, help_text=_('Saved scope details. Field is populated on scope deletion.'))
    parents = models.ManyToManyField('PriceEstimate', related_name='children', help_text=_('Price estimate parents'))

    total = models.FloatField(default=0, help_text=_('Predicted price for scope for current month.'))
    consumed = models.FloatField(default=0, help_text=_('Price for resource until now.'))

    month = models.PositiveSmallIntegerField(validators=[MaxValueValidator(12), MinValueValidator(1)])
    year = models.PositiveSmallIntegerField()

    objects = managers.PriceEstimateManager('scope')

    tracker = FieldTracker()

    class Meta:
        unique_together = ('content_type', 'object_id', 'month', 'year',)

    def __str__(self):
        name = self.get_scope_name() if self.scope else self.details.get('name')
        return '%s for %s-%s %.2f' % (name, self.year, self.month, self.total)

    @classmethod
    @lru_cache(maxsize=1)
    def get_estimated_models(cls):
        return (
            structure_models.ResourceMixin.get_all_models() +
            structure_models.ServiceProjectLink.get_all_models() +
            structure_models.Service.get_all_models() +
            [structure_models.ServiceSettings] +
            [structure_models.Project, structure_models.Customer]
        )

    def get_parents(self):  # For DescendantMixin
        return self.parents.all()

    def get_children(self):  # For DescendantMixin
        return self.children.all()

    def get_log_fields(self):  # For LoggableMixin
        return 'uuid', 'scope', 'total', 'consumed'

    def is_resource_estimate(self):
        return issubclass(self.content_type.model_class(), structure_models.ResourceMixin)

    def get_previous(self):
        """ Get estimate for the same scope for previous month. """
        month, year = (self.month - 1, self.year) if self.month != 1 else (12, self.year - 1)
        return PriceEstimate.objects.get(scope=self.scope, month=month, year=year)

    def create_ancestors(self):
        """ Create price estimates for scope ancestors if they do not exist """
        if not isinstance(self.scope, core_models.DescendantMixin):
            return
        scope_parents = self.scope.get_parents()
        for scope_parent in scope_parents:
            parent, created = PriceEstimate.objects.get_or_create(scope=scope_parent, month=self.month, year=self.year)
            self.parents.add(parent)
            if created:
                parent.create_ancestors()

    def init_details(self):
        """ Initialize price estimate details based on its scope """
        self.details = {
            'name': self.get_scope_name(),
            'description': getattr(self.scope, 'description', ''),
        }
        if hasattr(self.scope, 'backend_id'):
            self.details['backend_id'] = self.scope.backend_id
        if self.is_resource_estimate():
            self.details['service_settings_name'] = self.scope.service_project_link.service.settings.name
            self.details['project_name'] = self.scope.service_project_link.project.name
        self.save(update_fields=['details'])

    def update_total(self, update_ancestors=True, raise_exception=False):
        """ Re-calculate price of resource and its ancestors for the whole month,
            based on its configuration and consumption details.
        """
        self._check_is_updatable()
        new_total = self._get_price(self.consumption_details.consumed_in_month)
        diff = new_total - self.total
        with transaction.atomic():
            self.total = new_total
            self.save(update_fields=['total'])
            if update_ancestors:
                self.update_ancestors_total(diff, raise_exception=raise_exception)

    def update_ancestors_total(self, diff, raise_exception=False):
        for ancestor in self.get_ancestors():
            ancestor.total += diff
            ancestor.save(update_fields=['total'])

    def update_consumed(self):
        """ Re-calculate price of resource until now. Does not update ancestors. """
        self._check_is_updatable()
        self.consumed = self._get_price(self.consumption_details.consumed_until_now)
        self.save(update_fields=['consumed'])

    @classmethod
    def create_historical(cls, resource, configuration, date):
        """ Create price estimate and consumption details backdating.

            Method assumes that resource had given configuration from given date
            to the end of the month.
        """
        price_estimate = cls.objects.create(scope=resource, month=date.month, year=date.year)
        price_estimate.create_ancestors()
        # configuration is updated directly because we want to avoid recalculation
        # of consumed items based on current time.
        details = ConsumptionDetails(
            price_estimate=price_estimate,
            configuration=configuration,
            last_update_time=date,
        )
        details.save()
        price_estimate.update_total()
        return price_estimate

    def _get_price(self, consumed):
        """ Calculate price estimate for scope depends on consumed data and price list items.
            Map each consumable to price list item and multiply price its price by time of usage.
        """
        price_list_items = PriceListItem.get_for_resource(self.scope)
        consumables_prices = {(item.item_type, item.key): item.minute_rate for item in price_list_items}
        total = 0
        for consumable_item, usage in consumed.items():
            try:
                total += consumables_prices[(consumable_item.item_type, consumable_item.key)] * usage
            except KeyError:
                logger.debug('Price list item for consumable "%s" does not exist.' % consumable_item)
        return total

    def _check_is_updatable(self):
        """ Raise error if price estimate does not have consumption details or
            does not belong to resource
        """
        if not ConsumptionDetails.objects.filter(price_estimate=self).exists():
            raise EstimateUpdateError('Cannot update consumed for price estimate that does not have consumption details.')
        if not self.is_resource_estimate() or not self.scope:
            raise EstimateUpdateError('Cannot update consumed for price estimate that is not related to resource.')

    def get_scope_name(self):
        if self.scope:
            if isinstance(self.scope, (structure_models.ServiceProjectLink, structure_models.Service)):
                # We need to display some meaningful name for SPL.
                return str(self.scope)
            else:
                return self.scope.name
        else:
            return self.details.get('name')

    def collect_children(self):
        """
        Recursively collect children estimates. Returns generator.
        """
        for child in self.children.filter():
            yield child
            for grandchild in child.collect_children():
                yield grandchild

    @staticmethod
    def update_resource_estimate(resource, new_configuration, raise_exception=False):
        """ Create or update price estimate for resource based on its current configuration """
        price_estimate, created = PriceEstimate.objects.get_or_create_current(scope=resource)
        if created:
            price_estimate.create_ancestors()
        consumption_details, _ = ConsumptionDetails.objects.get_or_create(price_estimate=price_estimate)
        is_updated = consumption_details.update_configuration(new_configuration)
        if is_updated:
            price_estimate.update_total(raise_exception=raise_exception)
        return price_estimate


class ConsumptionDetailUpdateError(Exception):
    pass


class ConsumptionDetailCalculateError(Exception):
    pass


class ConsumableItemsField(JSONField):
    """ Store consumable items and their usage as JSON.

        Represent data in format:
        {
            <ConsumableItem instance>: <usage>,
            <ConsumableItem instance>: <usage>,
            ...
        }
        Store data in format:
        [
            {
                "item_type": xx,
                "key": xx,
                "usage": xx,
            }
            ...
        ]
    """

    def to_python(self, value):
        value = super(ConsumableItemsField, self).to_python(value)
        if isinstance(value, list):
            value = self._deserialize(value)
        return value

    def get_prep_value(self, value):
        if not isinstance(value, dict):
            raise TypeError('ConsumableItemsField value should be dict. Received: %s' % value)
        if any([not isinstance(item, ConsumableItem) for item in value]):
            raise TypeError('ConsumableItemsField keys should be instances of ConsumableItem class.')

        return super(ConsumableItemsField, self).get_prep_value(self._serialize(value))

    def _serialize(self, value):
        return [{'usage': usage, 'item_type': item.item_type, 'key': item.key}
                for item, usage in value.items()]

    def _deserialize(self, serialized_value):
        return {ConsumableItem(item['item_type'], item['key']): item['usage'] for item in serialized_value}


class ConsumptionDetails(core_models.UuidMixin, TimeStampedModel):
    """ Resource consumption details per month.

        Warning! Use method "update_configuration" to update configurations,
        do not update them manually.
    """
    price_estimate = models.OneToOneField(PriceEstimate, related_name='consumption_details', on_delete=models.CASCADE)
    configuration = ConsumableItemsField(default=dict, help_text=_('Current resource configuration.'))
    last_update_time = models.DateTimeField(help_text=_('Last configuration change time.'))
    consumed_before_update = ConsumableItemsField(
        default=dict, help_text=_('How many consumables were used by resource before last update.'))

    objects = managers.ConsumptionDetailsManager()

    class Meta:
        verbose_name = _('Consumption details')
        verbose_name_plural = _('Consumption details')

    def update_configuration(self, new_configuration):
        """ Save how much consumables were used and update current configuration.

            Return True if configuration changed.
        """
        if new_configuration == self.configuration:
            return False
        now = timezone.now()
        if now.month != self.price_estimate.month:
            raise ConsumptionDetailUpdateError('It is possible to update consumption details only for current month.')
        minutes_from_last_update = self._get_minutes_from_last_update(now)
        for consumable_item, usage in self.configuration.items():
            consumed_after_modification = usage * minutes_from_last_update
            self.consumed_before_update[consumable_item] = (
                self.consumed_before_update.get(consumable_item, 0) + consumed_after_modification)
        self.configuration = new_configuration
        self.last_update_time = now
        self.save()
        return True

    @property
    def consumed_in_month(self):
        """ How many resources were (or will be) consumed until end of the month """
        month_end = core_utils.month_end(datetime.date(self.price_estimate.year, self.price_estimate.month, 1))
        return self._get_consumed(month_end)

    @property
    def consumed_until_now(self):
        """ How many consumables were used by resource until now. """
        return self._get_consumed(timezone.now())

    def _get_consumed(self, time):
        """ How many consumables were (or will be) used by resource until given time. """
        minutes_from_last_update = self._get_minutes_from_last_update(time)
        if minutes_from_last_update < 0:
            raise ConsumptionDetailCalculateError('Cannot calculate consumption if time < last modification date.')
        _consumed = {}
        for consumable_item in set(list(self.configuration.keys()) + list(self.consumed_before_update.keys())):
            after_update = self.configuration.get(consumable_item, 0) * minutes_from_last_update
            before_update = self.consumed_before_update.get(consumable_item, 0)
            _consumed[consumable_item] = after_update + before_update
        return _consumed

    def _get_minutes_from_last_update(self, time):
        """ How much minutes passed from last update to given time """
        time_from_last_update = time - self.last_update_time
        return int(time_from_last_update.total_seconds() / 60)


class AbstractPriceListItem(models.Model):
    class Meta:
        abstract = True

    value = models.DecimalField("Hourly rate", default=0, max_digits=13, decimal_places=7)
    units = models.CharField(max_length=255, blank=True)

    @property
    def monthly_rate(self):
        return '%0.2f' % (self.value * core_utils.hours_in_month())

    @property
    def minute_rate(self):
        return float(self.value) / 60


class DefaultPriceListItem(core_models.UuidMixin, core_models.NameMixin, AbstractPriceListItem):
    """
    Default price list item for all resources of supported service types.

    It is fetched from cost tracking backend.
    Field "name" represents how price item will be represented for user.
    """
    item_type = models.CharField(max_length=255, help_text=_('Type of price list item. Examples: storage, flavor.'))
    key = models.CharField(
        max_length=255, help_text=_('Key that corresponds particular consumable. Example: name of flavor.'))
    resource_content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType, default=None)
    tracker = FieldTracker()

    def __str__(self):
        return 'Price list item %s: %s = %s for %s' % (self.name, self.key, self.value, self.resource_content_type)

    class Meta:
        unique_together = ('key', 'item_type', 'resource_content_type')

    @property
    def resource_type(self):
        cls = self.resource_content_type.model_class()
        if cls:
            return SupportedServices.get_name_for_model(cls)

    @classmethod
    def get_consumable_items_pretty_names(cls, resource_content_type, consumable_items):
        query = Q()
        for consumable_item in consumable_items:
            query |= (Q(item_type=consumable_item.item_type) & Q(key=consumable_item.key))
        price_list_items = cls.objects.filter(query, resource_content_type=resource_content_type)
        return {ConsumableItem(item_type, key): name
                for item_type, key, name in price_list_items.values_list('item_type', 'key', 'name')}

    @classmethod
    def init_from_registered_resources(cls):
        created_items = []
        with transaction.atomic():
            for resource_class in CostTrackingRegister.registered_resources:
                resource_content_type = ContentType.objects.get_for_model(resource_class)
                for consumable_item in CostTrackingRegister.get_consumable_items(resource_class):
                    price_list_item, created = cls._create_or_update_default_price_list_item(
                        resource_content_type, consumable_item)
                    if created:
                        created_items.append(price_list_item)
        return created_items

    @classmethod
    def _create_or_update_default_price_list_item(cls, resource_content_type, consumable_item):
        default_item, created = DefaultPriceListItem.objects.update_or_create(
            resource_content_type=resource_content_type,
            item_type=consumable_item.item_type,
            key=consumable_item.key,
            defaults={'units': consumable_item.units},
        )
        if created:
            default_item.value = consumable_item.default_price
            default_item.name = consumable_item.name
            default_item.save()
        return default_item, created


class PriceListItem(core_models.UuidMixin, AbstractPriceListItem):
    """
    Price list item related to private service.
    It is entered manually by customer owner.
    """
    # Generic key to service
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ContentType)
    object_id = models.PositiveIntegerField()
    service = GenericForeignKey('content_type', 'object_id')
    objects = managers.PriceListItemManager('service')
    default_price_list_item = models.ForeignKey(on_delete=models.CASCADE, to=DefaultPriceListItem)

    class Meta:
        unique_together = ('content_type', 'object_id', 'default_price_list_item')

    @property
    def resource_type(self):
        return self.default_price_list_item.resource_type

    @property
    def key(self):
        return self.default_price_list_item.key

    @property
    def item_type(self):
        return self.default_price_list_item.item_type

    def clean(self):
        if not self.service:
            raise ValidationError(_('Service is not defined.'))

        if SupportedServices.is_public_service(self.service):
            raise ValidationError(_('Public service does not support price list items.'))

        resource = self.default_price_list_item.resource_content_type.model_class()
        valid_resources = SupportedServices.get_related_models(self.service)['resources']

        if resource not in valid_resources:
            raise ValidationError(_('Service does not support required content type.'))

    @staticmethod
    def get_for_resource(resource):
        """ Get list of all price list items that should be used for resource.

            If price list item is defined for service - return it, otherwise -
            return default price list item.
        """
        resource_content_type = ContentType.objects.get_for_model(resource)
        default_items = set(DefaultPriceListItem.objects.filter(resource_content_type=resource_content_type))
        service = resource.service_project_link.service
        items = set(PriceListItem.objects.filter(
            default_price_list_item__in=default_items, service=service).select_related('default_price_list_item'))
        rewrited_defaults = set([i.default_price_list_item for i in items])
        return items | (default_items - rewrited_defaults)



from django.contrib import admin

from waldur_core.structure import admin as structure_admin
from .models import AzureService, AzureServiceProjectLink, VirtualMachine


admin.site.register(VirtualMachine, structure_admin.VirtualMachineAdmin)
admin.site.register(AzureService, structure_admin.ServiceAdmin)
admin.site.register(AzureServiceProjectLink, structure_admin.ServiceProjectLinkAdmin)



from csv import DictWriter

from django.core.management.base import BaseCommand, CommandError

from ... import models


class Command(BaseCommand):
    help_text = 'Export catalog of Amazon images.'
    args = '[ami_catalog.csv]'

    def handle(self, *args, **options):
        if len(args) == 0:
            raise CommandError('AMI catalog filename is not specified.')

        with open(args[0], 'w') as csvfile:
            writer = DictWriter(csvfile, fieldnames=('name', 'region', 'backend_id'))
            writer.writeheader()
            rows = [dict(name=image.name, region=image.region.name, backend_id=image.backend_id)
                    for image in models.Image.objects.all()]
            writer.writerows(rows)



import logging

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from waldur_core.core.models import StateMixin

from . import tasks, models

logger = logging.getLogger(__name__)


def notify_create_user(sender, instance, password, created=False, **kwargs):
    transaction.on_commit(lambda: tasks.notify_create_user.delay(instance.id,
                                                                 password,
                                                                 instance.settings.backend_url))


def delete_catalog_when_cluster_is_deleted(sender, instance, **kwargs):
    content_type = ContentType.objects.get_for_model(instance)
    models.Catalog.objects.filter(content_type=content_type, object_id=instance.id).delete()


def delete_node_if_related_instance_has_been_deleted(sender, instance, **kwargs):
    try:
        content_type = ContentType.objects.get_for_model(instance)
        node = models.Node.objects.get(object_id=instance.id, content_type=content_type)
        backend = node.cluster.get_backend()
        backend.delete_node(node)
    except ObjectDoesNotExist:
        pass


def delete_cluster_if_all_related_nodes_have_been_deleted(sender, instance, **kwargs):
    node = instance

    if node.cluster.state == models.Cluster.States.DELETING and not node.cluster.node_set.count():
        backend = node.cluster.get_backend()
        backend.delete_cluster(node.cluster)


def set_error_state_for_node_if_related_instance_deleting_is_failed(sender, instance, created=False, **kwargs):
    if created:
        return

    try:
        content_type = ContentType.objects.get_for_model(instance)
        node = models.Node.objects.get(object_id=instance.id, content_type=content_type)
    except ObjectDoesNotExist:
        return

    if instance.tracker.has_changed('state') and instance.state == StateMixin.States.ERRED:
        node.state = models.Node.States.ERRED
        node.error_message = 'Deleting related VM has failed.'
        node.save()


def set_error_state_for_cluster_if_related_node_deleting_is_failed(sender, instance, created=False, **kwargs):
    node = instance

    if created:
        return

    if node.tracker.has_changed('state') and node.state == models.Node.States.ERRED:
        if node.cluster.state == models.Cluster.States.DELETING:
            node.cluster.state = models.Cluster.States.ERRED
            node.cluster.error_message = 'Deleting one or a more nodes have failed.'
            node.cluster.save()



from datetime import timedelta

from waldur_core.core import WaldurExtension


class PayPalExtension(WaldurExtension):

    class Settings:
        WALDUR_PAYPAL = {
            'ENABLED': False,
            'BACKEND': {
                'mode': 'sandbox',  # either 'live' or 'sandbox'
                'client_id': '',
                'client_secret': '',
                'currency_name': 'USD',
            },
            'STALE_PAYMENTS_LIFETIME': timedelta(weeks=1)
        }

    @staticmethod
    def django_app():
        return 'waldur_paypal'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        from celery.schedules import crontab
        return {
            'debit-customers': {
                'task': 'waldur_paypal.DebitCustomers',
                'schedule': crontab(hour=0, minute=30),
                'args': (),
            },
            'payments-cleanup': {
                'task': 'waldur_paypal.PaymentsCleanUp',
                'schedule': timedelta(hours=24),
                'args': (),
            },
            'send-invoices': {
                'task': 'waldur_paypal.SendInvoices',
                'schedule': timedelta(hours=24),
                'args': (),
            }
        }



from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from djangosaml2.conf import get_config
from djangosaml2.utils import available_idps
from saml2.attribute_converter import ac_factory
from saml2.mdstore import InMemoryMetaData, MetaDataFile, name as get_idp_name
from saml2.s_utils import UnknownSystemEntity

from . import models


def load_providers():
    metadata = {}
    for filename in settings.WALDUR_AUTH_SAML2['idp_metadata_local']:
        mdf = MetaDataFile(ac_factory(), filename)
        mdf.load()
        metadata.update(mdf.items())
    return metadata


def sync_providers():
    providers = load_providers()

    current_idps = list(models.IdentityProvider.objects.all().only('url', 'pk'))
    backend_urls = set(providers.keys())

    stale_idps = set(idp.pk for idp in current_idps if idp.url not in backend_urls)
    models.IdentityProvider.objects.filter(pk__in=stale_idps).delete()

    existing_urls = set(idp.url for idp in current_idps)

    for url, metadata in providers.items():
        name = get_idp_name(metadata)
        if not name:
            # It is expected that every provider has name. For corner cases check entity_id
            name = metadata.get('entity_id')
            if not name:
                # Skip invalid identity provider
                continue
        if url in existing_urls:
            # Skip identity provider if its url is already in the database
            continue
        models.IdentityProvider.objects.create(url=url, name=name, metadata=metadata)

    for provider in models.IdentityProvider.objects.all().iterator():
        backend_metadata = providers.get(provider.url)
        if backend_metadata and provider.metadata != backend_metadata:
            provider.metadata = backend_metadata
            provider.save()


def is_valid_idp(value):
    remote_providers = available_idps(get_config()).keys()
    return value in remote_providers or models.IdentityProvider.objects.filter(url=value).exists()


def get_idp_sso_supported_bindings(idp_entity_id, config):
    try:
        return config.metadata.service(idp_entity_id, 'idpsso_descriptor', 'single_sign_on_service').keys()
    except (UnknownSystemEntity, AttributeError):
        return []


class DatabaseMetadataLoader(InMemoryMetaData):

    def load(self, *args, **kwargs):
        # Skip default parsing because data is not stored in file
        pass

    def __getitem__(self, item):
        try:
            return models.IdentityProvider.objects.get(url=item).metadata
        except ObjectDoesNotExist:
            raise KeyError



from waldur_core.core.tasks import ErrorStateTransitionTask

from .models import Volume


class SetInstanceErredTask(ErrorStateTransitionTask):
    """ Mark instance as erred and delete resources that were not created. """

    def execute(self, instance):
        super(SetInstanceErredTask, self).execute(instance)

        # delete volume if it were not created on backend,
        # mark as erred if creation was started, but not ended,
        volume = instance.volume_set.first()
        if volume.state == Volume.States.CREATION_SCHEDULED:
            volume.delete()
        elif volume.state == Volume.States.OK:
            pass
        else:
            volume.set_erred()
            volume.save(update_fields=['state'])



# Generated by Django 2.2.9 on 2020-02-21 21:17

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0007_changeemailrequest'),
    ]

    operations = [
        migrations.AddField(
            model_name='changeemailrequest',
            name='uuid',
            field=waldur_core.core.fields.UUIDField(),
        ),
    ]



from django.core.validators import MinValueValidator
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers as rf_serializers
from rest_framework import exceptions as rf_exceptions

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers
from waldur_core.structure.permissions import _has_owner_access
from waldur_freeipa import models as freeipa_models

from . import models


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        core_serializers.RequiredFieldsMixin,
                        structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_FIELDS = {
        'username': '',
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'hostname': _('Hostname or IP address of master node'),
        'port': '',
        'use_sudo': _('Set to true to activate privilege escalation'),
        'gateway': _('Hostname or IP address of gateway node'),
        'default_account': _('Default SLURM account for user'),
        'batch_service': _('Batch service, SLURM or MOAB')
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.SlurmService
        required_fields = ('hostname', 'username', 'batch_service')
        extra_field_options = {
            'username': {
                'default_value': 'root',
            },
            'use_sudo': {
                'default_value': False,
            },
            'default_account': {
                'required': True,
            },
        }


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.SlurmServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'slurm-detail'},
        }


class AllocationSerializer(structure_serializers.BaseResourceSerializer,
                           core_serializers.AugmentedSerializerMixin):
    service = rf_serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='slurm-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = rf_serializers.HyperlinkedRelatedField(
        view_name='slurm-spl-detail',
        queryset=models.SlurmServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    username = rf_serializers.SerializerMethodField()
    gateway = rf_serializers.SerializerMethodField()
    batch_service = rf_serializers.ReadOnlyField()
    homepage = rf_serializers.ReadOnlyField(
        source='service_project_link.service.settings.homepage')

    def get_username(self, allocation):
        request = self.context['request']
        try:
            profile = freeipa_models.Profile.objects.get(user=request.user)
            return profile.username
        except freeipa_models.Profile.DoesNotExist:
            return None

    def get_gateway(self, allocation):
        options = allocation.service_project_link.service.settings.options
        return options.get('gateway') or options.get('hostname')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Allocation
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'cpu_limit', 'cpu_usage',
            'gpu_limit', 'gpu_usage',
            'ram_limit', 'ram_usage',
            'deposit_limit', 'deposit_usage',
            'username', 'gateway',
            'is_active', 'batch_service', 'homepage'
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'cpu_usage', 'gpu_usage', 'ram_usage', 'is_active',
            'deposit_limit', 'deposit_usage',
        )
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'slurm-allocation-detail'},
            cpu_limit={'validators': [MinValueValidator(0)]},
            gpu_limit={'validators': [MinValueValidator(0)]},
            ram_limit={'validators': [MinValueValidator(0)]},
        )

    def validate(self, attrs):
        attrs = super(AllocationSerializer, self).validate(attrs)
        # Skip validation on update
        if self.instance:
            return attrs

        spl = attrs['service_project_link']
        user = self.context['request'].user
        if not _has_owner_access(user, spl.project.customer):
            raise rf_exceptions.PermissionDenied(
                _('You do not have permissions to create allocation for given project.')
            )
        return attrs


class AllocationUsageSerializer(rf_serializers.HyperlinkedModelSerializer):
    full_name = rf_serializers.ReadOnlyField(source='user.full_name')

    class Meta:
        model = models.AllocationUsage
        fields = ('allocation', 'year', 'month',
                  'username', 'user', 'full_name',
                  'cpu_usage', 'ram_usage', 'gpu_usage', 'deposit_usage')
        extra_kwargs = {
            'allocation': {
                'lookup_field': 'uuid',
                'view_name': 'slurm-allocation-detail',
            },
            'user': {
                'lookup_field': 'uuid',
                'view_name': 'user-detail',
            }
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-02-22 09:00
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0011_unique_together'),
    ]

    operations = [
        migrations.AlterField(
            model_name='attachment',
            name='backend_id',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='comment',
            name='backend_id',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='backend_id',
            field=models.CharField(max_length=255, null=True),
        ),
    ]



from django import forms
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import admin as core_admin

from . import models


class PresetAdminForm(forms.ModelForm):
    class Meta:
        widgets = {
            'ram': core_admin.GBtoMBWidget(),
            'storage': core_admin.GBtoMBWidget(),
        }


class PresetAdmin(admin.ModelAdmin):
    def ram_gb(self, obj):
        return '%s GB' % core_admin.GBtoMBWidget().format_value(obj.ram)

    ram_gb.admin_order_field = 'ram'
    ram_gb.short_description = _('RAM')

    def storage_gb(self, obj):
        return '%s GB' % core_admin.GBtoMBWidget().format_value(obj.storage)

    storage_gb.admin_order_field = 'storage'
    storage_gb.short_description = _('Storage')

    form = PresetAdminForm
    base_model = models.Preset
    list_display = ('category', 'variant', 'name', 'ram_gb', 'cores', 'storage_gb')
    list_filter = ('category',)


class PresetInline(admin.TabularInline):
    form = PresetAdminForm
    model = models.Preset
    extra = 1


class CategoryAdmin(admin.ModelAdmin):
    inlines = (PresetInline,)


class DeploymentPlanItem(admin.TabularInline):
    model = models.DeploymentPlanItem
    extra = 1
    fields = ('preset', 'quantity')


class DeploymentPlanAdmin(admin.ModelAdmin):
    inlines = (DeploymentPlanItem,)
    search_fields = ('name',)
    list_display = ('name', 'project')


admin.site.register(models.Category, CategoryAdmin)
admin.site.register(models.Preset, PresetAdmin)
admin.site.register(models.DeploymentPlan, DeploymentPlanAdmin)



from django.core.management.base import BaseCommand
from reversion.models import Version

from waldur_core.quotas.models import Quota


class Command(BaseCommand):
    help = "Delete quotas versions duplicates."

    def handle(self, *args, **options):
        self.stdout.write('Collecting duplicates...')
        duplicates = sum([self.get_quota_duplicate_versions(quota) for quota in Quota.objects.all()], [])
        self.stdout.write('...Done')

        if not duplicates:
            self.stdout.write('No duplicates were found. Congratulations!')
        else:
            self.stdout.write('There are %s duplicates for quotas versions.' % len(duplicates))
            while True:
                delete = input('  Do you want to delete them? [Y/n]:') or 'y'
                if delete.lower() not in ('y', 'n'):
                    self.stdout.write('  Please enter letter "y" or "n"')
                else:
                    delete = delete.lower() == 'y'
                    break
            if delete:
                for duplicate in duplicates:
                    duplicate.delete()
                self.stdout.write('All duplicates were deleted.')
            else:
                self.stdout.write('Duplicates were not deleted.')

    def get_quota_duplicate_versions(self, quota):
        versions = Version.objects.get_for_object(quota).order_by('revision__date_created')
        if not versions:
            return []
        duplicates = []
        last_version = versions[0]
        for version in versions[1:]:
            next_version = version
            if self.are_versions_equal(quota, last_version, next_version):
                duplicates.append(next_version)
            else:
                last_version = next_version
        return duplicates

    def are_versions_equal(self, quota, v1, v2):
        o1 = v1._object_version.object
        o2 = v2._object_version.object
        return all([getattr(o1, f) == getattr(o2, f) for f in quota.get_version_fields()])



# Generated by Django 2.2.7 on 2019-12-11 10:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0001_squashed_0046'),
    ]

    operations = [
        migrations.AddField(
            model_name='volumetype',
            name='is_default',
            field=models.BooleanField(default=False),
        ),
    ]



from django.conf import settings

from waldur_core.core import utils
from waldur_core.core.models import StateMixin

from . import tasks


def detect_vm_coordinates(sender, instance, name, source, target, **kwargs):
    # Check if geolocation is enabled
    if not settings.WALDUR_CORE.get('ENABLE_GEOIP', True):
        return

    # VM already has coordinates
    if instance.latitude is not None and instance.longitude is not None:
        return

    if target == StateMixin.States.OK:
        tasks.detect_vm_coordinates.delay(utils.serialize_instance(instance))



import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from rest_framework import status, serializers
from rest_framework.reverse import reverse

from waldur_core.structure import models as structure_models, SupportedServices
from waldur_mastermind.common import utils as common_utils
from waldur_mastermind.marketplace import models, signals
from waldur_mastermind.marketplace.utils import validate_limits

logger = logging.getLogger(__name__)


def get_spl_url(spl_model_class, order_item):
    """
    Find service project link URL for specific service settings and marketplace order.
    """
    service_settings = order_item.offering.scope

    service_settings_type = spl_model_class._meta.app_config.service_name

    if not isinstance(service_settings, structure_models.ServiceSettings) or \
            service_settings.type != service_settings_type:
        raise serializers.ValidationError('Offering has invalid scope. Service settings object is expected.')

    project = order_item.order.project

    try:
        spl = spl_model_class.objects.get(
            project=project,
            service__settings=service_settings,
            service__customer=project.customer,
        )
        return reverse('{}-detail'.format(spl.get_url_name()), kwargs={'pk': spl.pk})
    except ObjectDoesNotExist:
        raise serializers.ValidationError('Project does not have access to the service.')


def copy_attributes(fields, order_item):
    """
    Copy valid attributes from order item.
    """
    payload = dict()
    for field in fields:
        if field in order_item.attributes:
            payload[field] = order_item.attributes.get(field)
    return payload


class BaseOrderItemProcessor:
    def __init__(self, order_item):
        self.order_item = order_item

    def process_order_item(self, user):
        """
        This method receives user object and creates plugin's resource corresponding
        to provided order item. It is called after order has been approved.
        """
        raise NotImplementedError()

    def validate_order_item(self, request):
        """
        This method receives request object, and raises
        validation error if provided order item is invalid.
        It is called after order has been created but before it is submitted.
        """
        raise NotImplementedError()


class AbstractCreateResourceProcessor(BaseOrderItemProcessor):
    def process_order_item(self, user):
        scope = self.send_request(user)

        with transaction.atomic():
            resource = models.Resource(
                project=self.order_item.order.project,
                offering=self.order_item.offering,
                plan=self.order_item.plan,
                limits=self.order_item.limits,
                attributes=self.order_item.attributes,
                name=self.order_item.attributes.get('name') or '',
                scope=scope,
            )
            resource.init_cost()
            resource.save()
            resource.init_quotas()
            self.order_item.resource = resource
            self.order_item.save(update_fields=['resource'])

    def send_request(self, user):
        """
        This method should send request to backend.
        """
        raise NotImplementedError


class CreateResourceProcessor(AbstractCreateResourceProcessor):
    """
    This class implements order processing using internal API requests.

    Order item validation flow looks as following:
    1) Convert order item to HTTP POST request data expected by DRF serializer.
    2) Pass request data to serializer and check if data is valid.

    Order item processing flow looks as following:
    1) Convert order item to HTTP POST request data expected by DRF serializer.
    2) Issue internal API request to DRF viewset.
    3) Extract Django model for created resource from HTTP response.
    4) Create marketplace resource object from order item and plugin resource.
    5) Store link from order item to the resource.

    Therefore this class implements template method design pattern.
    """

    def validate_order_item(self, request):
        post_data = self.get_post_data()
        serializer_class = self.get_serializer_class()
        context = {'request': request, 'skip_permission_check': True}
        serializer = serializer_class(data=post_data, context=context)
        serializer.is_valid(raise_exception=True)

    def send_request(self, user):
        post_data = self.get_post_data()
        view = self.get_viewset().as_view({'post': 'create'})
        response = common_utils.create_request(view, user, post_data)
        if response.status_code != status.HTTP_201_CREATED:
            raise serializers.ValidationError(response.data)

        return self.get_scope_from_response(response)

    def get_serializer_class(self):
        """
        This method should return DRF serializer class which
        validates request data to provision new resources.
        """
        raise NotImplementedError

    def get_viewset(self):
        """
        This method should return DRF viewset class which
        processes request to provision new resources.
        """
        raise NotImplementedError

    def get_post_data(self):
        """
        This method converts order item to request data expected by DRF serializer.
        """
        raise NotImplementedError

    def get_scope_from_response(self, response):
        """
        This method extracts Django model from response returned by DRF viewset.
        """
        raise NotImplementedError


class AbstractUpdateResourceProcessor(BaseOrderItemProcessor):
    def is_update_limit_order_item(self):
        if 'old_limits' in self.order_item.attributes.keys():
            return True

    def validate_order_item(self, request):
        if self.is_update_limit_order_item():
            validate_limits(self.order_item.limits, self.order_item.offering)
            return

        post_data = self.get_post_data()
        serializer_class = self.get_serializer_class()
        context = {'request': request, 'skip_permission_check': True}
        serializer = serializer_class(data=post_data, context=context)
        serializer.is_valid(raise_exception=True)

    def process_order_item(self, user):
        """We need to overwrite process order item because two cases exist:
        a switch of a plan and a change of limits."""
        if self.is_update_limit_order_item():
            try:
                # self.update_limits_process method can execute not is_async
                # because in this case an order has got only one order item.
                self.update_limits_process(user)
            except NotImplementedError:
                self.order_item.set_state_erred()
                self.order_item.save(update_fields=['state'])
                logger.warning('An update of limits has been called. '
                               'But update limits process for the plugin has not been implemented. '
                               'Order item ID: %s, Plugin: %s.',
                               self.order_item.id, self.order_item.offering.type)
            except Exception as e:
                signals.limit_update_failed.send(
                    sender=self.order_item.resource.__class__,
                    order_item=self.order_item,
                    error_message=str(e),
                )
            else:
                signals.limit_update_succeeded.send(
                    sender=self.order_item.resource.__class__,
                    order_item=self.order_item,
                )
            return

        resource = self.get_resource()
        if not resource:
            raise serializers.ValidationError('Resource is not found.')

        self.send_request(user)
        self.order_item.resource.set_state_updating()
        self.order_item.resource.save(update_fields=['state'])

    def send_request(self, user):
        """
        This method should send request to backend.
        """
        raise NotImplementedError

    def get_resource(self):
        """
        This method should return related resource of order item.
        """
        return self.order_item.resource.scope

    def update_limits_process(self, user):
        """
        This method implements limits update processing.
        """
        raise NotImplementedError


class UpdateResourceProcessor(AbstractUpdateResourceProcessor):
    def send_request(self, user):
        view = self.get_view()
        payload = self.get_post_data()
        response = common_utils.create_request(view, user, payload)
        if response.status_code != status.HTTP_202_ACCEPTED:
            raise serializers.ValidationError(response.data)

    def get_serializer_class(self):
        """
        This method should return DRF serializer class which
        validates request data to update existing resource.
        """
        raise NotImplementedError

    def get_view(self):
        """
        This method should return DRF viewset class which
        processes request to change existing resource.
        """
        raise NotImplementedError

    def get_post_data(self):
        """
        This method converts order item to request data expected by DRF serializer.
        """
        raise NotImplementedError


class AbstractDeleteResourceProcessor(BaseOrderItemProcessor):
    def validate_order_item(self, request):
        pass

    def get_resource(self):
        """
        This method should return related resource of order item.
        """
        return self.order_item.resource.scope

    def send_request(self, user, resource):
        """
        This method should send request to backend.
        """
        raise NotImplementedError

    def process_order_item(self, user):
        resource = self.get_resource()
        if not resource:
            raise serializers.ValidationError('Resource is not found.')

        done = self.send_request(user, resource)

        if done:
            with transaction.atomic():
                self.order_item.resource.set_state_terminated()
                self.order_item.resource.save(update_fields=['state'])

                self.order_item.state = models.OrderItem.States.DONE
                self.order_item.save(update_fields=['state'])
        else:
            with transaction.atomic():
                self.order_item.resource.set_state_terminating()
                self.order_item.resource.save(update_fields=['state'])


class DeleteResourceProcessor(AbstractDeleteResourceProcessor):
    viewset = NotImplementedError

    def send_request(self, user, resource):
        view = self.get_viewset().as_view({'delete': 'destroy'})
        delete_attributes = self.order_item.attributes
        response = common_utils.delete_request(view, user, uuid=resource.uuid.hex, query_params=delete_attributes)
        if response.status_code not in (status.HTTP_204_NO_CONTENT, status.HTTP_202_ACCEPTED):
            raise serializers.ValidationError(response.data)
        return response.status_code == status.HTTP_204_NO_CONTENT

    def get_viewset(self):
        """
        This method should return DRF viewset class which
        processes request to delete existing resource.
        """
        return self.viewset


class BaseCreateResourceProcessor(CreateResourceProcessor):
    """
    Abstract base class to adapt resource provisioning endpoints to marketplace API.
    It is assumed that resource model and serializer uses service project link.
    """
    viewset = NotImplementedError
    fields = NotImplementedError

    def get_viewset(self):
        return self.viewset

    def get_fields(self):
        """
        Get list of valid attribute names to be copied from order item to
        resource provisioning request.
        """
        return self.fields

    def get_resource_model(self):
        """
        Get resource model used by viewset from its queryset.
        """
        return self.get_viewset().queryset.model

    def get_spl_model(self):
        """
        Get service project link model used by resource model using service registry.
        """
        return SupportedServices.get_related_models(self.get_resource_model())['service_project_link']

    def get_serializer_class(self):
        """
        Use create_serializer_class if it is defined. Otherwise fallback to standard serializer class.
        """
        viewset = self.get_viewset()
        return getattr(viewset, 'create_serializer_class', None) or getattr(viewset, 'serializer_class')

    def get_post_data(self):
        order_item = self.order_item
        return dict(
            service_project_link=get_spl_url(self.get_spl_model(), order_item),
            **copy_attributes(self.get_fields(), order_item)
        )

    def get_scope_from_response(self, response):
        return self.get_resource_model().objects.get(uuid=response.data['uuid'])



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-13 10:39
from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
    ]

    operations = [
        migrations.CreateModel(
            name='SlurmPackage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('product_code', models.CharField(blank=True, max_length=30)),
                ('article_code', models.CharField(blank=True, max_length=30)),
                ('cpu_price', models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price for CPU hour')),
                ('gpu_price', models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price for GPU hour')),
                ('ram_price', models.DecimalField(decimal_places=10, default=0, max_digits=14, validators=[django.core.validators.MinValueValidator(Decimal('0'))], verbose_name='Price for GB RAM')),
                ('service_settings', models.OneToOneField(limit_choices_to={'type': 'SLURM'}, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'SLURM package',
                'verbose_name_plural': 'SLURM packages',
            }
        ),
    ]



from rest_framework.permissions import BasePermission, SAFE_METHODS


class IsAdminOrReadOnly(BasePermission):
    def has_permission(self, request, view):
        return (
            request.method in SAFE_METHODS or
            (request.user.is_authenticated and request.user.is_staff)
        )


class ActionsPermission(BasePermission):
    """
    Allow to define custom permission checks for all actions together and each action separately.

    It is possible to define permissions checks in next way:
     - view.safe_methods_permissions - list of checks for all safe methods (GET, OPTIONS, HEAD).
     - view.unsafe_methods_permissions - list of checks for all unsafe methods (POST, PUT, PATCH, DELETE).
     - view.<action>_extra_permissions - list of action extra permissions. Backend will check
                                         view level permissions and extra_permissions together.
     - view.<action>_permissions- list of all view permissions. Backend will not check view level
                                  permissions if action permissions are defined.

    Example. Define action level permissions:

        def is_staff(request, view, obj=None):
            if not request.user.is_staff:
                raise PermissionDenied('User has to be staff to perform this action.')

        class MyView(...):
            permission_classes = (ActionsPermission,)
            ...
            def action(...):
                ...

            action_permissions = [is_staff]  # action will be available only for staff

    Example. Define view level permissions and additional permissions for
    action:

        def is_staff(request, view, obj=None):
            if not request.user.is_staff:
                raise PermissionDenied('User has to be staff to perform this action.')

        def has_civil_number(request, view, obj=None):
            if not request.user.civil_number:
                raise PermissionDenied('User has to have civil number to perform this action.')

        class MyView(...):
            permission_classes = (ActionsPermission,)
            # only user with civil number will have access to all unsafe actions
            unsafe_methods_permissions = [has_civil_number]
            ...
            @decorators.action(detail=True, method='POST')
            def action(...):
                ...

            action_extra_permissions = [is_staff]  # only staff user with civil numbers will have access to action
    """

    def get_permission_checks(self, request, view):
        """
        Get permission checks that will be executed for current action.
        """
        if view.action is None:
            return []
        # if permissions are defined for view directly - use them.
        if hasattr(view, view.action + '_permissions'):
            return getattr(view, view.action + '_permissions')
        # otherwise return view-level permissions + extra view permissions
        extra_permissions = getattr(view, view.action + 'extra_permissions', [])
        if request.method in SAFE_METHODS:
            return getattr(view, 'safe_methods_permissions', []) + extra_permissions
        else:
            return getattr(view, 'unsafe_methods_permissions', []) + extra_permissions

    def has_permission(self, request, view):
        for check in self.get_permission_checks(request, view):
            check(request, view)
        return True

    def has_object_permission(self, request, view, obj):
        for check in self.get_permission_checks(request, view):
            check(request, view, obj)
        return True



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:02
from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('playbook_jobs', '0005_add_job_deleting_status'),
    ]

    operations = [
        migrations.AlterField(
            model_name='job',
            name='arguments',
            field=waldur_core.core.fields.JSONField(blank=True, default=dict, null=True),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core.fields import NaturalChoiceField
from waldur_core.core.serializers import RestrictedSerializerMixin
from waldur_core.logging import models, loggers


class EventSerializer(RestrictedSerializerMixin,
                      serializers.ModelSerializer):
    context = serializers.JSONField(read_only=True)

    class Meta:
        model = models.Event
        fields = ('uuid', 'created', 'event_type', 'message', 'context')


class BaseHookSerializer(serializers.HyperlinkedModelSerializer):
    author_uuid = serializers.ReadOnlyField(source='user.uuid')
    hook_type = serializers.SerializerMethodField()

    class Meta:
        model = models.BaseHook

        fields = (
            'url', 'uuid', 'is_active', 'author_uuid',
            'event_types', 'event_groups', 'created', 'modified',
            'hook_type'
        )

        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def get_fields(self):
        """
        When static declaration is used, event type choices are fetched too early -
        even before all apps are initialized. As a result, some event types are missing.
        When dynamic declaration is used, all valid event types are available as choices.
        """
        fields = super(BaseHookSerializer, self).get_fields()
        fields['event_types'] = serializers.MultipleChoiceField(
            choices=loggers.get_valid_events(), required=False)
        fields['event_groups'] = serializers.MultipleChoiceField(
            choices=loggers.get_event_groups_keys(), required=False)
        return fields

    def create(self, validated_data):
        validated_data['user'] = self.context['request'].user
        return super(BaseHookSerializer, self).create(validated_data)

    def validate(self, attrs):
        if not self.instance and 'event_types' not in attrs and 'event_groups' not in attrs:
            raise serializers.ValidationError(_('Please specify list of event_types or event_groups.'))

        if 'event_groups' in attrs:
            events = list(attrs.get('event_types', []))
            groups = list(attrs.get('event_groups', []))
            events = sorted(set(loggers.expand_event_groups(groups)) | set(events))

            attrs['event_types'] = events
            attrs['event_groups'] = groups

        elif 'event_types' in attrs:
            attrs['event_types'] = list(attrs['event_types'])

        return attrs

    def get_hook_type(self, hook):
        raise NotImplementedError


class SummaryHookSerializer(serializers.Serializer):

    def to_representation(self, instance):
        serializer = self.get_hook_serializer(instance.__class__)
        return serializer(instance, context=self.context).data

    def get_hook_serializer(self, cls):
        for serializer in BaseHookSerializer.__subclasses__():
            if serializer.Meta.model == cls:
                return serializer
        raise ValueError('Hook serializer for %s class is not found' % cls)


class WebHookSerializer(BaseHookSerializer):
    content_type = NaturalChoiceField(models.WebHook.ContentTypeChoices.CHOICES, required=False)

    class Meta(BaseHookSerializer.Meta):
        model = models.WebHook
        fields = BaseHookSerializer.Meta.fields + ('destination_url', 'content_type')

    def get_hook_type(self, hook):
        return 'webhook'


class PushHookSerializer(BaseHookSerializer):
    type = NaturalChoiceField(models.PushHook.Type.CHOICES)

    class Meta(BaseHookSerializer.Meta):
        model = models.PushHook
        fields = BaseHookSerializer.Meta.fields + ('type', 'device_id', 'token', 'device_manufacturer', 'device_model')

    def get_hook_type(self, hook):
        return 'pushhook'


class EmailHookSerializer(BaseHookSerializer):

    class Meta(BaseHookSerializer.Meta):
        model = models.EmailHook
        fields = BaseHookSerializer.Meta.fields + ('email', )

    def get_hook_type(self, hook):
        return 'email'



""" Defines how to optimize price for OpenStackTenant instances """
import collections

from django.contrib.contenttypes.models import ContentType
from rest_framework import serializers as rf_serializers

from waldur_core.cost_tracking import models as cost_tracking_models
from waldur_openstack.openstack_tenant import (
    apps as ot_apps, models as ot_models, serializers as ot_serializers, cost_tracking as ot_cost_tracking)

from .. import optimizers, register, serializers


OptimizedPreset = collections.namedtuple(
    'OptimizedPreset', ('preset', 'flavor', 'quantity', 'price', 'flavor_price', 'storage_price'))


OptimizedOpenStackTenant = optimizers.namedtuple_with_defaults(
    'OptimizedOpenStack',
    field_names=optimizers.OptimizedService._fields + ('optimized_presets',),
    default_values=optimizers.OptimizedService._defaults,
)


class OpenStackTenantOptimizer(optimizers.Optimizer):
    """ Find the cheapest OpenStackTenant flavor for each preset. """
    HOURS_IN_DAY = 24

    def _get_service_price_item(self, service, resource_content_type, item_type, key):
        default_item = cost_tracking_models.DefaultPriceListItem.objects.get(
            resource_content_type=resource_content_type, item_type=item_type, key=key)
        try:
            return cost_tracking_models.PriceListItem.objects.get(
                default_price_list_item=default_item, service=service)
        except cost_tracking_models.PriceListItem.DoesNotExist:
            return default_item

    def _get_flavor_price(self, service, flavor):
        instance_content_type = ContentType.objects.get_for_model(ot_models.Instance)
        try:
            item = self._get_service_price_item(
                service=service,
                resource_content_type=instance_content_type,
                item_type=ot_cost_tracking.InstanceStrategy.Types.FLAVOR,
                key=flavor.name,
            )
        except cost_tracking_models.DefaultPriceListItem.DoesNotExist:
            raise optimizers.OptimizationError('Price is not defined for flavor "%s".' % flavor.name)
        return item.value * self.HOURS_IN_DAY

    def _get_cheapest_flavor(self, service, suitable_flavors):
        priced_flavors = [(flavor, self._get_flavor_price(service, flavor)) for flavor in suitable_flavors]
        cheapest_flavor, price = min(priced_flavors, key=lambda x: x[1])
        return cheapest_flavor, price

    def _get_storage_price(self, service, storage):
        volume_content_type = ContentType.objects.get_for_model(ot_models.Volume)
        try:
            item = self._get_service_price_item(
                service=service,
                resource_content_type=volume_content_type,
                item_type=ot_cost_tracking.VolumeStrategy.Types.STORAGE,
                key=ot_cost_tracking.VolumeStrategy.Keys.STORAGE,
            )
        except cost_tracking_models.DefaultPriceListItem.DoesNotExist:
            raise optimizers.OptimizationError('Storage price is not defined.')
        return item.value * self.HOURS_IN_DAY

    def optimize(self, deployment_plan, service):
        optimized_presets = []
        price = 0
        for item in deployment_plan.items.all():
            preset = item.preset
            suitable_flavors = ot_models.Flavor.objects.filter(
                cores__gte=preset.cores, ram__gte=preset.ram, settings=service.settings)
            if not suitable_flavors:
                preset_as_str = '%s (cores: %s, ram %s MB, storage %s MB)' % (
                    preset.name, preset.cores, preset.ram, preset.storage)
                raise optimizers.OptimizationError(
                    'It is impossible to create an instance for preset %s. It is too big.' % preset_as_str)

            flavor, flavor_price = self._get_cheapest_flavor(service, suitable_flavors)
            storage_price = self._get_storage_price(service, preset.storage)
            preset_price = flavor_price + storage_price
            optimized_presets.append(OptimizedPreset(
                preset=preset,
                flavor=flavor,
                quantity=item.quantity,
                flavor_price=flavor_price,
                storage_price=storage_price,
                price=preset_price * item.quantity,
            ))
            price += preset_price * item.quantity
        return OptimizedOpenStackTenant(price=price, service=service, optimized_presets=optimized_presets)


register.Register.register_optimizer(ot_apps.OpenStackTenantConfig.service_name, OpenStackTenantOptimizer)


class OptimizedPresetSerializer(rf_serializers.Serializer):
    flavor = ot_serializers.FlavorSerializer()
    preset = serializers.PresetSerializer()
    quantity = rf_serializers.IntegerField()
    flavor_price = rf_serializers.DecimalField(max_digits=22, decimal_places=10)
    storage_price = rf_serializers.DecimalField(max_digits=22, decimal_places=10)
    price = rf_serializers.DecimalField(max_digits=22, decimal_places=10)


class OptimizedOpenStackTenantSerializer(serializers.OptimizedServiceSerializer):
    service = rf_serializers.HyperlinkedRelatedField(
        view_name='openstacktenant-detail',
        lookup_field='uuid',
        read_only=True,
    )
    optimized_presets = OptimizedPresetSerializer(many=True)


register.Register.register_serializer(ot_apps.OpenStackTenantConfig.service_name, OptimizedOpenStackTenantSerializer)



from . import views


def register_in(router):
    router.register(r'slurm', views.SlurmServiceViewSet, basename='slurm')
    router.register(r'slurm-service-project-link', views.SlurmServiceProjectLinkViewSet,
                    basename='slurm-spl')
    router.register(r'slurm-allocation', views.AllocationViewSet, basename='slurm-allocation')
    router.register(r'slurm-allocation-usage', views.AllocationUsageViewSet, basename='slurm-allocation-usage')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-10-18 16:42
import django_fsm
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('playbook_jobs', '0004_add_spl_key_subnet_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='error_message',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='job',
            name='state',
            field=django_fsm.FSMIntegerField(choices=[(5, 'Creation Scheduled'), (6, 'Creating'), (1, 'Update Scheduled'), (2, 'Updating'), (7, 'Deletion Scheduled'), (8, 'Deleting'), (3, 'OK'), (4, 'Erred')], default=5),
        ),
    ]
